"""
Package with functionalities for generating customizable html,
that can be used to test selectors and workflows.

Classes
-------
- `AttributeGenerator` - Generator for creating html attributes.
- `TagGenerator` - Generator for creating html tags.
- `ChoiceTemplate` - Template for randomly selecting from a list of choices.
- `RandomTemplate` - Template for generating a random string of ASCII letters and digits.
- `BaseTemplate` - Base for user-defined templates.
"""

from .generators import AttributeGenerator, TagGenerator
from .generators.templates import ChoiceTemplate, RandomTemplate
from .generators.templates.base import BaseTemplate

__all__ = [
    "AttributeGenerator",
    "TagGenerator",
    "ChoiceTemplate",
    "RandomTemplate",
    "BaseTemplate",
]



"""
Selector for finding elements based on XPath.
`lxml` adapter, that allows any supported XPath expressions
to be used with other `soupsavvy` components.

XPath search is not natively supported by `bs4`, finding element is delegated
to `lxml` library, and corresponding `bs4` elements are selected.

Classes
-------
- XPathSelector
"""

from itertools import islice
from typing import Any, Optional, Union

from bs4 import Tag

try:
    from lxml import html
    from lxml.etree import XPath, XPathSyntaxError
    from lxml.html.soupparser import _convert_tree as to_lxml
except ImportError as e:
    raise ImportError(
        "`soupsavvy.selectors.xpath` requires `lxml` package to be installed."
    ) from e

import warnings

import soupsavvy.exceptions as exc
from soupsavvy.base import SoupSelector
from soupsavvy.utils.selector_utils import TagIterator


class XPathSelector(SoupSelector):
    """
    Selector for finding elements based on XPath expressions
    based on `lxml` implementation.

    Examples
    --------
    >>> selector = XPathSelector("//p[@class='menu']")
    ... selector.find(soup)

    Examples
    --------
    >>> from lxml.etree import XPath
    ... selector = XPathSelector(XPath("//p[@class='menu']", smart_strings=False))
    ... selector.find(soup)

    Expressions must target elements, not attributes or text content,
    otherwise, warning is raised and no results are found.

    Examples
    --------
    >>> selector = XPathSelector("//div//@href")
    ... selector.find(soup)
    None

    Notes
    -----
    Equality check includes only xpath expression, as lxml `XPath` object
    does not implement more specific `__eq__` method.
    """

    def __init__(self, xpath: Union[str, XPath]) -> None:
        """
        Initializes `XPathSelector` with a given XPath expression.

        Parameters
        ----------
        xpath : str | lxml.etree.XPath
            String representing of xpath expression or compiled `XPath` object.
            It needs to target elements, not attributes or text content.

        Raises
        ------
        InvalidXPathSelector
            If the provided XPath string cannot be compiled into `XPath` object.
        """
        if not isinstance(xpath, XPath):
            try:
                xpath = XPath(xpath)
            except XPathSyntaxError as e:
                raise exc.InvalidXPathSelector(f"Parsing XPath '{xpath}' failed") from e

        self.xpath = xpath

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        element: html.HtmlElement = to_lxml(tag, None)  # type: ignore
        matches: list = self.xpath(element)  # type: ignore

        if matches and all(
            not isinstance(match, html.HtmlElement) for match in matches
        ):
            warnings.warn(
                "XPath expression does not target elements, no results will be found.",
                UserWarning,
            )
            return []

        # this approach is possible due to DFS traversal
        # strongly based on assumption, that both iterators are in sync
        lxml_iterator = (
            element.iterdescendants(None) if recursive else element.iterchildren(None)
        )
        soup_iterator = TagIterator(tag, recursive=recursive)

        return list(
            islice(
                (
                    soup
                    for soup, element in zip(soup_iterator, lxml_iterator)
                    if element in matches
                ),
                limit,
            )
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, XPathSelector):
            return False

        return self.xpath.path == other.xpath.path

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.xpath!r})"



"""
Module for base classes used across the package.
Introduces another layer of abstraction for operations and selectors.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, overload

from bs4 import NavigableString, Tag
from typing_extensions import deprecated

import soupsavvy.exceptions as exc
import soupsavvy.selectors.namespace as ns
from soupsavvy.interfaces import Comparable, Executable, TagSearcher

if TYPE_CHECKING:
    from soupsavvy.operations.general import OperationPipeline
    from soupsavvy.operations.selection_pipeline import SelectionPipeline
    from soupsavvy.selectors.combinators import (
        AncestorCombinator,
        ChildCombinator,
        DescendantCombinator,
        NextSiblingCombinator,
        ParentCombinator,
        SubsequentSiblingCombinator,
    )
    from soupsavvy.selectors.logical import AndSelector, SelectorList, XORSelector


def check_selector(x: Any, message: Optional[str] = None) -> SoupSelector:
    """
    Checks if provided object is a valid `soupsavvy` selector.
    Checks for instance of `SoupSelector` and raises an exception if not.
    Returns provided object if fulfills the condition for convenience.

    Parameters
    ----------
    x : Any
        Any object to be validated as correct selector.
    message : str, optional
        Custom message to be displayed in case of raising an exception.
        By default None, which results in default message.

    Raises
    ------
    NotSoupSelectorException
        If provided object is not an instance of `SoupSelector`.
    """
    message = message or f"Object {x} is not an instance of {SoupSelector.__name__}."

    if not isinstance(x, SoupSelector):
        raise exc.NotSoupSelectorException(message)

    return x


def check_operation(x: Any, message: Optional[str] = None) -> BaseOperation:
    """
    Checks if provided object is a valid `soupsavvy` operation.
    Checks for instance of `BaseOperation` and raises an exception if not.
    Returns provided object if fulfills the condition for convenience.

    Parameters
    ----------
    x : Any
        Any object to be validated as correct operation.
    message : str, optional
        Custom message to be displayed in case of raising an exception.
        By default None, which results in default message.

    Raises
    ------
    NotOperationException
        If provided object is not an instance of `BaseOperation`.
    """
    message = message or f"Object {x} is not an instance of {BaseOperation.__name__}."

    if not isinstance(x, BaseOperation):
        raise exc.NotOperationException(message)

    return x


class SoupSelector(TagSearcher, Comparable):
    """
    Base class for all `soupsavvy` selectors, that define declarative search procedure
    of searching for elements in BeautifulSoup Tag.

    Selectors can be combined with other selectors to create search procedures.
    They can be chained with operations to extract and transform the data.

    Methods
    -------
    - find(tag: Tag, strict: bool = False, recursive: bool = True) -> Optional[Tag]

        Finds first element matching selector in provided tag if found.
        If no element is found, returns None by default,
        or raises an exception if `strict` mode is enabled.
        Additionally `recursive` parameter can be set to search only direct children.

    - find_all(tag: Tag, recursive: bool = True, limit: Optional[int] = None) -> list[Tag]

        Finds all elements matching selector in provided tag and returns them in a list.
        Additionally `limit` and `recursive` parameters can be set.

    Notes
    -----
    - Specific selector inheriting from this class, need to implement:
        - `find_all` method that returns a list of matching elements in bs4.Tag.
        - `__eq__` method to compare two selectors for equality.
    - Optionally `find` method can be implemented to return first matching element,
    but, by default, it uses `find_all` under the hood.
    """

    @overload
    def find(
        self,
        tag: Tag,
        strict: Literal[False] = ...,
        recursive: bool = ...,
    ) -> Optional[Tag]: ...

    @overload
    def find(
        self,
        tag: Tag,
        strict: Literal[True] = ...,
        recursive: bool = ...,
    ) -> Tag: ...

    @overload
    def find(
        self,
        tag: Tag,
        strict: bool = ...,
        recursive: bool = ...,
    ) -> Optional[Tag]: ...

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Optional[Tag]:
        """
        Finds the first matching element in provided `Tag`.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to search within.
        strict : bool, optional
            If `True`, raises an exception if tag was not found in markup,
            if `False` and tag was not found, returns `None`.
            Value of this parameter does not affect behavior if element
            was successfully found. By default `False`.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.

        Returns
        -------
        Tag | None
            First `bs4.Tag` object matching selector or `None` if none matching.

        Raises
        ------
        TagNotFoundException
            If strict parameter is set to `True` and none matching tag was found.
        NavigableStringException
            If `bs4.NavigableString` was returned by selector.
        """
        result = self._find(tag, recursive=recursive)

        if result is None:
            if strict:
                raise exc.TagNotFoundException("Tag was not found in markup.")
            return None

        if isinstance(result, NavigableString):
            raise exc.NavigableStringException(
                f"NavigableString was returned for {result} string search, "
                "invalid operation for SoupSelector find."
            )

        return result

    @abstractmethod
    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        """
        Finds all elements matching selector in provided `Tag`.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to search within.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.
        limit : int, optional
            Specifies maximum number of elements to return.
            By default `None`, all found elements are returned.

        Returns
        -------
        list[Tag]
            List of `bs4.Tag` objects matching selector. If none found, the list is empty.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    def _find(self, tag: Tag, recursive: bool = True) -> ns.FindResult:
        """
        Returns an object that is a result of tag search.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to search within.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.

        Returns
        -------
        NavigableString | Tag | None:
            Result of selector search within provided `bs4.Tag`.
        """
        elements = self.find_all(tag, recursive=recursive, limit=1)
        return elements[0] if elements else None

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Checks self and other object for equality.

        This method is abstract and must be implemented by all selectors.

        Calling `find` or `find_all` methods on selectors that are equal
        should return the same results, provided the same parameters.

        Example
        -------
        >>> selector1 = TypeSelector("div")
        ... selector2 = TypeSelector("div")
        ... selector1 == selector2
        True
        ... selector1.find(tag) == selector2.find(tag)
        True
        """

        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    @overload
    def __or__(self, x: SoupSelector) -> SelectorList: ...

    @overload
    def __or__(self, x: BaseOperation) -> SelectionPipeline: ...

    def __or__(
        self, x: Union[SoupSelector, BaseOperation]
    ) -> Union[SelectorList, SelectionPipeline]:
        """
        Overrides `__or__` method called also by pipe operator `|`.

        Syntactical Sugar for logical disjunction, that creates `SelectorList`,
        when combining two `SoupSelector` objects, or `SelectionPipeline`,
        when combining `SoupSelector` with `BaseOperation`.

        Parameters
        ----------
        x : SoupSelector | BaseOperation
            `SoupSelector` object to be combined into new `SelectorList`
            object as `SelectorList(this, other)` or `BaseOperation` to bew chained
            with it as `SelectionPipeline(this, other)`.

        Returns
        -------
        SelectorList
            Union of two `SoupSelector` objects, that can be used to get all elements
            that match at least one of the elements in the Union.
        SelectionPipeline
            Pipeline of `SoupSelector` and `BaseOperation` objects that can be used
            to extract and transform the data from the elements.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`
            or `BaseOperation`.
        """
        from soupsavvy.base import BaseOperation
        from soupsavvy.operations.selection_pipeline import SelectionPipeline
        from soupsavvy.selectors.logical import SelectorList

        if isinstance(x, BaseOperation):
            return SelectionPipeline(selector=self, operation=x)

        message = (
            f"Bitwise OR not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__} or {BaseOperation.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, SelectorList):
            args = [*self.selectors, x]
            # return new SelectorList with updated steps
            return SelectorList(*args)

        return SelectorList(self, x)

    def __invert__(self) -> SoupSelector:
        """
        Overrides `__invert__` method called also by tilde operator `~`.
        Syntactical Sugar for bitwise NOT operator, that creates `NotSelector` instance
        matching everything that is not matched by this `SoupSelector`.

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be negated into new `NotSelector` as `NotSelector(this)`.

        Returns
        -------
        NotSelector
            `NotSelector(this)`.
        SoupSelector
            Any `SoupSelector` object in case of directly inverting
            `NotSelector` to get original `SoupSelector`.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.logical import NotSelector

        return NotSelector(self)

    def __and__(self, x: SoupSelector) -> AndSelector:
        """
        Overrides `__and_`_ method called also by ampersand operator `&`.
        Syntactical Sugar for bitwise AND operator, that creates `AndSelector` instance
        matching everything that is matched by both `SoupSelector` objects.

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `AndSelector`
            object as `AndSelector(this, other)`.

        Returns
        -------
        AndSelector
            `AndSelector(this, other)`.

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.logical import AndSelector

        message = (
            f"Bitwise AND not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, AndSelector):
            args = [*self.selectors, x]
            # return new AndSelector with updated steps
            return AndSelector(*args)

        return AndSelector(self, x)

    def __gt__(self, x: SoupSelector) -> ChildCombinator:
        """
        Overrides `__gt__` method called also by greater-than operator `>`.
        Syntactical Sugar for greater-than operator, that creates `ChildCombinator`.

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `ChildCombinator`
            object as `ChildCombinator(this, other)`.

        Returns
        -------
        ChildCombinator
            `ChildCombinator(this, other)`
        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import ChildCombinator

        message = (
            f"GT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, ChildCombinator):
            args = [*self.selectors, x]
            # return new ChildCombinator with updated steps
            return ChildCombinator(*args)

        return ChildCombinator(self, x)

    def __lt__(self, x: SoupSelector) -> ParentCombinator:
        """
        Overrides `__lt__` method called also by less-than operator `<`.
        Syntactical Sugar for less-than operator, that creates `ParentCombinator` instance.

        Example
        -------
        >>> TypeSelector("a") < TypeSelector("div")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `ParentCombinator`
            object as `ParentCombinator(this, other)`.

        Returns
        -------
        ParentCombinator
            `ParentCombinator(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import ParentCombinator

        message = (
            f"LT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, ParentCombinator):
            args = [*self.selectors, x]
            # return new ParentCombinator with updated steps
            return ParentCombinator(*args)

        return ParentCombinator(self, x)

    def __add__(self, x: SoupSelector) -> NextSiblingCombinator:
        """
        Overrides `__add__` method called also by plus operator `+`.
        Syntactical Sugar for addition operator, that creates `NextSiblingCombinator`
        instance.

        Example
        -------
        >>> TypeSelector("div") + TypeSelector("a")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new` NextSiblingCombinator`
            object as `NextSiblingCombinator(this, other)`.

        Returns
        -------
        NextSiblingCombinator
            `NextSiblingCombinator(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import NextSiblingCombinator

        message = (
            f"ADD operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, NextSiblingCombinator):
            args = [*self.selectors, x]
            # return new NextSiblingCombinator with updated steps
            return NextSiblingCombinator(*args)

        return NextSiblingCombinator(self, x)

    def __mul__(self, x: SoupSelector) -> SubsequentSiblingCombinator:
        """
        Overrides `__mul__` method called also by multiplication operator `*`.
        Syntactical Sugar for multiplication operator, that creates
        `SubsequentSiblingCombinator` instance.

        Example
        -------
        >>> TypeSelector("div") * TypeSelector("a")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `SubsequentSiblingCombinator`
            object as `SubsequentSiblingCombinator(this, other)`.

        Returns
        -------
        SubsequentSiblingCombinator
            `SubsequentSiblingCombinator(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import SubsequentSiblingCombinator

        message = (
            f"MUL operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, SubsequentSiblingCombinator):
            args = [*self.selectors, x]
            # return new SubsequentSiblingCombinator with updated steps
            return SubsequentSiblingCombinator(*args)

        return SubsequentSiblingCombinator(self, x)

    def __rshift__(self, x: SoupSelector) -> DescendantCombinator:
        """
        Overrides `__rshift__` method called also by right shift operator `>>`.
        Syntactical Sugar for right shift operator,
        that creates `DescendantCombinator` instance.

        Example
        -------
        >>> TypeSelector("div") >> TypeSelector("a")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `DescendantCombinator`
            object as `DescendantCombinator(this, other)`.

        Returns
        -------
        DescendantCombinator
            `DescendantCombinator(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import DescendantCombinator

        message = (
            f"RIGHT SHIFT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, DescendantCombinator):
            args = [*self.selectors, x]
            # return new DescendantCombinator with updated steps
            return DescendantCombinator(*args)

        return DescendantCombinator(self, x)

    def __lshift__(self, x: SoupSelector) -> AncestorCombinator:
        """
        Overrides `__lshift__` method called also by left shift operator `<<`.
        Syntactical Sugar for left-shift operator, that creates `AncestorCombinator` instance.

        Example
        -------
        >>> TypeSelector("a") << TypeSelector("div")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `AncestorCombinator`
            object as `AncestorCombinator(this, other)`.

        Returns
        -------
        AncestorCombinator
            `AncestorCombinator(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.combinators import AncestorCombinator

        message = (
            f"LEFT SHIFT operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, AncestorCombinator):
            args = [*self.selectors, x]
            # return new AncestorCombinator with updated steps
            return AncestorCombinator(*args)

        return AncestorCombinator(self, x)

    def __xor__(self, x: SoupSelector) -> XORSelector:
        """
        Overrides `__xor__` method called also by `xor` (caret) operator `^`.
        Syntactical Sugar `xor` operator, that creates `XORSelector` instance.

        Example
        -------
        >>> TypeSelector("a") ^ TypeSelector("div")

        Parameters
        ----------
        x : SoupSelector
            `SoupSelector` object to be combined into new `XORSelector`
            object as `XORSelector(this, other)`.

        Returns
        -------
        XORSelector
            `XORSelector(this, other)`

        Raises
        ------
        NotSoupSelectorException
            If provided object is not an instance of `SoupSelector`.
        """
        from soupsavvy.selectors.logical import XORSelector

        message = (
            f"XOR operator not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {SoupSelector.__name__}."
        )
        check_selector(x, message=message)

        if isinstance(self, XORSelector):
            args = [*self.selectors, x]
            # return new XORSelector with updated steps
            return XORSelector(*args)

        return XORSelector(self, x)


class SelectableCSS(ABC):
    """
    Interface for selectors, that can clearly and unambiguously defined css selector,
    used to search for elements, that matches the same tags as find methods.

    Notes
    -----
    To implement SelectableCSS interface, child class must implement:
    - 'selector' property, which return a string representing css selector.
    """

    @property
    @abstractmethod
    def css(self) -> str:
        """Returns string representing element css selector."""
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this property."
        )

    @property
    @deprecated("'selector' property is deprecated, use 'css' instead.")
    def selector(self) -> str:
        """Returns string representing element css selector."""
        return self.css


class CompositeSoupSelector(SoupSelector):
    """
    Interface for selectors consisting of multiple selectors.

    Notes
    -----
    To implement `CompositeSoupSelector` interface, child class must
    call its init method with provided selectors to set up the object.

    Attributes
    ----------
    selectors : list[SoupSelector]
        List of `SoupSelector` objects used for searching elements.
    """

    # if the order of selectors is not relevant - by default True
    COMMUTATIVE = True

    def __init__(self, selectors: Iterable[SoupSelector]) -> None:
        """
        Initializes composite selector object with provided selectors.
        Checks if all selectors are instances of `SoupSelector`.

        Parameters
        ----------
        selectors: Iterable[SoupSelector]
            Selectors used to search for elements.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of `SoupSelector`.
        """
        self._selectors = [check_selector(selector) for selector in selectors]

    @property
    def selectors(self) -> list[SoupSelector]:
        """
        Returns a list of selectors that composite selector consists of.

        Returns
        -------
        list[SoupSelector]
            List of `SoupSelector` objects used for searching elements.
        """
        return self._selectors

    def __eq__(self, other: object) -> bool:
        # check for CompositeSoupSelector type for type checking sake
        if not isinstance(other, CompositeSoupSelector):
            return False
        elif type(self) is not type(other):
            # checking for exact type match - isinstance(other, self.__class__)
            # when other is subclass of self.__class__ would call other.__eq__(self)
            # which is not desired behavior, as it returns False
            return False

        if not self.__class__.COMMUTATIVE:
            return self.selectors == other.selectors

        # if order is irrelevant, check if all selectors are covered by other
        # and vice versa - selector lists can be of different size as well
        return all(
            any(self_selector == other_selector for other_selector in other.selectors)
            for self_selector in self.selectors
        ) and all(
            any(other_selector == self_selector for self_selector in self.selectors)
            for other_selector in other.selectors
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(str, self.selectors))})"

    def __repr__(self) -> str:
        return str(self)


class BaseOperation(Executable, Comparable):
    """
    Base class for all `soupsavvy` operations.
    Operations are used to process the selection results from the soup,
    extract and transform the data.

    Operations can be chained together using the pipe operator '|'.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... operation = Operation(str.lower) | Operation(str.strip)
    ... operation.execute("  TEXT  ")
    'text'

    Operations can be combined with selectors to extract and transform
    target information.

    Example
    -------
    >>> from soupsavvy import TypeSelector
    ... from soupsavvy.operations import Operation, Text
    ... selector = TypeSelector("div") | Text() | Operation(int)
    ... selector.find(soup)
    42

    `BaseOperation` inherits from `Comparable` interface,
    `__eq__` method needs to be implemented in derived classes.
    """

    def execute(self, arg: Any) -> Any:
        """
        Execute the operation on the given argument and return the result.

        Parameters
        ----------
        arg : Any
            Argument to be processed by the operation.

        Returns
        -------
        Any
            Result of the operation.
        """
        try:
            return self._execute(arg)
        except exc.BreakOperationException:
            # break exception is propagated to the caller to handle
            raise
        except Exception as e:
            raise exc.FailedOperationExecution(
                f"Failed to execute operation: {e}"
            ) from e

    @abstractmethod
    def _execute(self, arg: Any) -> Any:
        """Internal method for executing the operation."""

        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface, "
            "and does not implement this method."
        )

    def __or__(self, x: Any) -> OperationPipeline:
        """
        Overrides `__or__` method called also by pipe operator '|'.
        Syntactical Sugar for logical disjunction, that creates `OperationPipeline` instance,
        which chains two operations together.

        Parameters
        ----------
        x : BaseOperation
            `BaseOperation` object to be combined into new `OperationPipeline`
            with this operation.

        Returns
        -------
        OperationPipeline
            `OperationPipeline(this, other)`.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        from soupsavvy.operations.general import OperationPipeline

        message = (
            f"Bitwise OR not supported for types {type(self)} and {type(x)}, "
            f"expected an instance of {BaseOperation.__name__}."
        )
        other = check_operation(x, message=message)
        return OperationPipeline(self, other)


class OperationSearcherMixin(BaseOperation, TagSearcher):
    """
    Mixin of `BaseOperation` and `TagSearcher` interfaces.
    Allows operations to be used as field searchers in model
    to perform operation directly on scope element.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Processes provided element and returns the result in a list.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        recursive : bool, optional
            Ignored, for consistency with interface.
        limit : int, optional
            Ignored, for consistency with interface.

        Returns
        -------
        list[Any]
            Result of applied operation on element in a list.
        """
        return [self.execute(tag)]

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Processes provided element and returns the result.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        strict : bool, optional
            Ignored, for consistency with interface.
        limit : int, optional
            Ignored, for consistency with interface.

        Returns
        -------
        Any
            Result of applied operation on element.
        """
        return self.execute(tag)



"""
Module with combinators, which are composite selectors, that define a relationship
between multiple selectors related to their position in the document.

Most of them are counterpart of CSS combinators, but others extend their functionality.

Classes
-------
- `ChildCombinator` - counterpart of CSS child combinator (>)
- `NextSiblingCombinator` - counterpart of CSS adjacent sibling combinator (+)
- `SubsequentSiblingCombinator` - counterpart of CSS subsequent sibling combinator (*)
- `DescentCombinator` - counterpart of CSS descendant combinator (" ")
- `ParentCombinator` - matches parent of preceding selector
- `AncestorCombinator` - matches ancestor of preceding selector

Notes
-----
For more information on CSS combinators, see:

https://developer.mozilla.org/en-US/docs/Learn/CSS/Building_blocks/Selectors/Combinators
"""

from abc import abstractmethod
from functools import reduce
from typing import Optional, Type

from bs4 import Tag
from typing_extensions import deprecated

from soupsavvy.base import CompositeSoupSelector, SoupSelector
from soupsavvy.selectors.logical import SelectorList as _SelectorList
from soupsavvy.selectors.relative import (
    RelativeAncestor,
    RelativeChild,
    RelativeDescendant,
    RelativeNextSibling,
    RelativeParent,
    RelativeSelector,
    RelativeSubsequentSibling,
)
from soupsavvy.utils.selector_utils import TagIterator, TagResultSet


@deprecated("`SelectorList` was moved to `soupsavvy.selectors.logical` module.")
class SelectorList(_SelectorList): ...


class BaseCombinator(CompositeSoupSelector):
    """
    Base class for all combinators, which are composite selectors,
    that defined a relationship between multiple selectors and apply it to search
    for elements in the document.
    """

    # order of selectors is relevant in context of results
    COMMUTATIVE = False

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes Combinator object with provided positional arguments.
        At least two `SoupSelector` object are required to create Combinator.

        Parameters
        ----------
        selectors: SoupSelector
            `SoupSelector` objects to match accepted as positional arguments.

        Notes
        -----
        Object can be initialized with more than two `SoupSelector` objects,
        which would be equal to chaining multiple combinators of the same type.

        For example, chaining child combinator in css:

        Example
        -------
        >>> div > a > span

        translated to `soupsavvy` would be:

        Example
        -------
        >>> ChildCombinator(TypeSelector("div"), TypeSelector("a"), TypeSelector("span"))

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    @property
    @abstractmethod
    def _selector(self) -> Type[RelativeSelector]:
        """
        Returns type of the relative selector that is used
        to perform a single step search in the combinator selector.

        Returns
        -------
        Type[RelativeSelector]
            Type of the relative selector that is used in the combinator.
            Selector instance of this type is initialized
            with each step in the combinator.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is a base class "
            "and does not implement '_selector' property."
        )

    def _find_first_step(
        self, step: SoupSelector, tag: Tag, recursive: bool
    ) -> TagResultSet:
        """
        Returns results of the first step in the combinator selector,
        given Tag object that is being searched and recursive behavior
        passed to find method by user.

        Parameters
        ----------
        tag: Tag
            Tag object that's being searched by first step in the combinator.
        recursive: bool
            Recursive behavior passed to find method by user.

        Returns
        -------
        TagResultSet
            Results of the first step in the combinator selector.
        """
        return TagResultSet(step.find_all(tag, recursive=recursive))

    def _order_results(
        self, results: TagResultSet, tag: Tag, recursive: bool
    ) -> TagResultSet:
        """
        Orders results of find_all method of the combinator selector, given
        initial Tag object that was passed to find_all method and recursive behavior.

        Parameters
        ----------
        results: TagResultSet
            Results of the combinator selector.
        tag: Tag
            Initial Tag object that was passed to find_all method.
        recursive: bool
            Recursive behavior passed to find method by user.

        Returns
        -------
        TagResultSet
            Ordered results of the combinator selector.
        """
        return TagResultSet(list(TagIterator(tag, recursive=True))) & results

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        results = TagResultSet()

        for i, step in enumerate(self.selectors):
            if i == 0:
                results |= self._find_first_step(
                    step=step, tag=tag, recursive=recursive
                )
                continue

            if not results:
                break

            selector = self._selector(step)
            results = TagResultSet(
                reduce(
                    list.__add__,
                    # each relative selector has defined recursive behavior
                    (selector.find_all(element) for element in results.fetch()),
                )
            )

        results = self._order_results(results=results, tag=tag, recursive=recursive)
        return results.fetch(limit)


class BaseAncestorCombinator(BaseCombinator):
    """
    Base class for ancestor combinators, that are specific type of combinators,
    unlike other combinators, they move up the tree of elements, rather than down
    after finding first step.

    - Elements that match first step can be found
    anywhere in the tree, regardless of `recursive` parameter.
    - Final results should contain only children of element if `recursive` is False.
    """

    def _find_first_step(
        self, step: SoupSelector, tag: Tag, recursive: bool
    ) -> TagResultSet:
        # always look for all descendants in first step
        return TagResultSet(step.find_all(tag, recursive=True))

    def _order_results(
        self, results: TagResultSet, tag: Tag, recursive: bool
    ) -> TagResultSet:
        # respect recursive parameter while ordering results
        return TagResultSet(list(TagIterator(tag, recursive=recursive))) & results


class ChildCombinator(BaseCombinator):
    """
    Counterpart of CSS child combinator.
    Represents the relationship between selectors, where every next matching element
    is a direct child of the previous one.

    Example
    -------
    >>> ChildCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that are direct children of 'div' elements.

    Example
    -------
    >>> <div class="widget"><a>Hello World</a></div> ✔️
    >>> <div class="widget"><span></span><a>Hello World</a></div> ✔️
    >>> <span class="widget"><a>Hello World</a></span> ❌
    >>> <div class="menu"><span>Hello World</span></div> ❌

    Object can be created as well by using `greater than` operator `>`
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") > TypeSelector("a")

    Which is equivalent to the first example.

    CSS counterpart can be represented as:

    Example
    -------
    >>> div > a { color: red; }

    Notes
    -----
    For more information on child combinator, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Child_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeChild


class NextSiblingCombinator(BaseCombinator):
    """
    Counterpart of CSS next sibling combinator.
    Represents the relationship between selectors, where every next matching element
    is a sibling immediately following the previous one.

    Example
    -------
    >>> NextSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that immediately follow 'div' elements,
    it means that both elements are children of the same parent element.

    Example
    -------
    >>> <div class="widget"></div><a>Hello World</a> ✔️
    >>> <div class="widget"><a>Hello World</a></div> ❌
    >>> <div class="widget"></div><span></span><a>Hello World</a> ❌

    Object can be created as well by using `plus` operator `+` on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") + TypeSelector("a")

    Which is equivalent to the first example.

    CSS counterpart can be represented as:

    Example
    -------
    >>> div + a

    Notes
    -----
    This is also known as the `adjacent sibling combinator` in CSS.
    For more information on next sibling combinator, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Next-sibling_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeNextSibling


class SubsequentSiblingCombinator(BaseCombinator):
    """
    Counterpart of CSS subsequent sibling combinator.
    Represents the relationship between selectors, where every next matching element
    is a sibling following the previous one, but not necessarily immediately.

    Example
    -------
    >>> SubsequentSiblingCombinator(TypeSelector("div"), TypeSelector("a"))

    matches all 'a' elements that follow 'div' elements.

    Example
    -------
    >>> <div class="widget"></div><a>Hello World</a> ✔️
    >>> <div class="widget"><span></span><a>Hello World</a></div> ✔️
    >>> <span class="widget"><a>Hello World</a></span> ❌
    >>> <a>Hello World</a><div class="menu"></div> ❌

    Object can be created as well by using `multiplication` operator `*`
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") * TypeSelector("a")

    CSS counterpart can be represented as:

    Example
    -------
    >>> div ~ a

    Notes
    -----
    This combinator is also known as `general sibling combinator` in CSS.
    For more information on subsequent sibling combinator, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Subsequent-sibling_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeSubsequentSibling


class DescendantCombinator(BaseCombinator):
    """
    Counterpart of CSS descendant combinator.
    Represents the relationship between selectors, where every next matching element
    is a descendant of the previous one.

    Example
    -------
    >>> DescentCombinator(TypeSelector("div"), ClassSelector("widget"))

    matches all descendants of 'div' element with 'widget' class.

    Example
    -------
    >>> <div><a class="widget"></a></div> ✔️
    >>> <div><div><a class="widget"></a></div></div> ✔️
    >>> <div><a id="widget"></a></div> ❌
    >>> <span><a class="widget"></a></span> ❌
    >>> <a class="widget"></a> ❌

    Object can be created as well by using `right shift` operator `>>`
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") >> ClassSelector("widget")

    CSS counterpart can be represented as:

    Example
    -------
    >>> div .widget

    Notes
    -----
    For more information on subsequent sibling combinator, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Descendant_combinator
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeDescendant


class ParentCombinator(BaseAncestorCombinator):
    """
    Defines a relationship between selectors, where every next matching element
    is a parent of the previous one.

    Example
    -------
    >>> ParentCombinator(TypeSelector("a"), TypeSelector("div"))

    The given selector matches all 'div' elements that are parents of 'a' elements.

    Example
    -------
    >>> <div><a href="/shop"></a></div> ✔️
    >>> <div><span><div><a href="/shop"></a></span></div> ❌
    >>> <span><a href="/shop"></a></span> ❌

    Object can be created as well by using `lt` operator `<` on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("a") < TypeSelector("div")

    Although this combinator does not have its counterpart in CSS, it can be represented as:

    Example
    -------
    >>> div:has(> a)
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeParent


class AncestorCombinator(BaseAncestorCombinator):
    """
    Defines a relationship between selectors, where every next matching element
    is an ancestor of the previous one.

    Example
    -------
    >>> AncestorCombinator(TypeSelector("a"), TypeSelector("div"))

    The given selector matches all 'div' elements that are ancestors of 'a' elements.

    Example
    -------
    >>> <div><span><a href="/shop"></a></span></div> ✔️
    >>> <div><a href="/shop"></a></div> ✔️
    >>> <div><span class="menu"></span>/div> ❌
    >>> <span><a class="menu"></span>/div> ❌

    Object can be created as well by using `left shift` operator `<<`
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("a") << TypeSelector("div")

    Although this combinator does not have its counterpart in CSS, it can be represented as:

    Example
    -------
    >>> div:has(a)
    """

    @property
    def _selector(self) -> Type[RelativeSelector]:
        return RelativeAncestor



"""
Package with operations used to post-process the results of selections.

Classes
-------
- `Operation` - User defined operation with any function.
- `Text` - Operation to extract text from the tag.
- `Href` - Operation to extract href attribute from the tag.
- `Parent` - Operation to extract parent tag of the tag.
- `Break` - Operation to break the pipeline execution.
- `IfElse` - Operation to control flow in the pipeline.
- `Continue` - Operation to skip the operation and continue with the next one.
- `SkipNone` - Operation to skip operation if input is None.
- `Suppress` - Operation to suppress exceptions and return None instead.
"""

from .conditional import Break, Continue, IfElse
from .general import Operation
from .soup import Href, Parent, Text
from .wrappers import SkipNone, Suppress

__all__ = [
    "Operation",
    "Text",
    "Href",
    "Parent",
    "Break",
    "IfElse",
    "Continue",
    "SkipNone",
    "Suppress",
]



"""
Module with `soupsavvy` interfaces used across the package.

- `Executable` - Interface for operations that can be executed on single argument.
- `Comparable` - Interface for objects that can be compared for equality.
- `TagSearcher` - Interface for objects that can search within `bs4.Tag`.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc


class Executable(ABC):
    """
    Interface for operations that can be executed on single argument.
    Derived classes must implement the `execute` method.
    """

    @abstractmethod
    def execute(self, arg: Any) -> Any:
        """Executes the operation on the given argument."""
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface "
            "and does not implement this method."
        )


class Comparable(ABC):
    """
    Interface for objects that can be compared for equality.
    Derived classes must implement the `__eq__` method.
    """

    @abstractmethod
    def __eq__(self, x: Any) -> bool:
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface "
            "and does not implement this method."
        )


# possible exceptions raised when TagSearcher fails
TagSearcherExceptions = (exc.FailedOperationExecution, exc.TagNotFoundException)


class TagSearcher(ABC):
    """
    Interface for objects that can search within `bs4.Tag`.
    Derived classes must implement the `find` and `find_all` methods,
    that process `bs4.Tag` object and return results.
    """

    @abstractmethod
    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Processes `bs4.Tag` object and returns result.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        strict : bool, optional
            If True, enforces results to be found in the tag, by default False.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.

        Returns
        -------
        Any
            Processed result from the tag.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface "
            "and does not implement this method."
        )

    @abstractmethod
    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Processes `bs4.Tag` object and returns list of results.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.
        limit : int, optional
            Specifies maximum number of results to return in a list.
            By default `None`, everything is returned.

        Returns
        -------
        list[Any]
            A list of results from processed tag.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} is an interface "
            "and does not implement this method."
        )



"""
Module with interface for generators.

`soupsavvy` generator is anything that generates content,
they should all inherit from this class and implement the 'generate' method.
"""

from abc import ABC, abstractmethod


class BaseGenerator(ABC):
    """
    Interface for all kind of `soupsavvy` generators.

    All components generating content should inherit from this class
    and implement the `generate` method to return the generated content.
    """

    @abstractmethod
    def generate(self) -> str:
        raise NotImplementedError(
            f"{self.__class__.__name__} is a base class and does not implement "
            "'generate' method."
        )



"""
Module for conditional operations to control flow in the pipeline.

Classes
-------
- `IfElse` - Executes different operations based on the condition.
- `Break` - Stops the pipeline execution.
- `Continue` - Skips the operation and continues with the next one.
"""

from collections.abc import Callable
from typing import Any

import soupsavvy.exceptions as exc
from soupsavvy.base import BaseOperation, OperationSearcherMixin, check_operation

Condition = Callable[[Any], bool]


class IfElse(OperationSearcherMixin):
    """
    Operation to control flow in the pipeline.
    Allows to execute different operations based on the condition.

    Example
    -------
    >>> from soupsavvy.operations import IfElse, Operation
    ... operation = IfElse(
    ...     lambda x: x == 0,
    ...     Operation(lambda x: None),
    ...     Operation(lambda x: x / 100),
    ... )
    ... operation.execute(0)
    None
    ... operation.execute(100)
    1

    Implements `TagSearcher` interface for convenience. It can conditionally
    apply operations to the tag and can be used as model field.

    Example
    -------
    >>> from soupsavvy.operations import IfElse, Operation, Text
    ... operation = IfElse(
    ...     lambda x: x.get("id") == "user",
    ...     Text(),
    ...     Href(),
    ... )
    ... operation.find(user_tag)
    username
    ... operation.find(other_tag)
    www.example.com
    """

    def __init__(
        self,
        condition: Condition,
        if_: BaseOperation,
        else_: BaseOperation,
    ) -> None:
        """
        Initializes `IfEls`e operation with condition and two operations.

        Parameters
        ----------
        condition : Callable[[Any], bool]
            Condition to check if the operation should be executed.
            If callable returns True, `if_` operation is executed, otherwise `else_`.
        if_ : BaseOperation
            Operation to execute if condition is fulfilled.
        else_ : BaseOperation
            Operation to execute if condition is not fulfilled.
        """
        self._condition = condition
        self._if_operation = check_operation(if_)
        self._else_operation = check_operation(else_)

    def _execute(self, arg: Any) -> Any:
        condition = self._condition(arg)
        operation = self._if_operation if condition else self._else_operation
        return operation.execute(arg)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IfElse):
            return False

        return (
            self._condition is other._condition
            and self._if_operation == other._if_operation
            and self._else_operation == other._else_operation
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(condition={self._condition}, "
            f"if_={self._if_operation}, else_={self._else_operation})"
        )

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._condition} ? {self._if_operation} : {self._else_operation})"
        )


class Break(OperationSearcherMixin):
    """
    Operation to break the pipeline execution and return the current result.
    Can be used in selection/operation pipelines with `IfElse` operation
    to conditionally stop the execution.

    Example
    -------
    >>> from soupsavvy.operations import Break, IfElse, Operation
    ... operation = IfElse(
    ...     lambda x: x == 0,
    ...     Break(),
    ...     Operation(lambda x: 100 / x),
    ... ) | Operation(lambda x: x + 1)
    ... operation.execute(0)
    0
    ... operation.execute(100)
    2

    If `Break` operation is executed, the pipeline will stop and return the result,
    so next operation is not executed.
    """

    def _execute(self, arg: Any) -> Any:
        raise exc.BreakOperationException(arg)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Break)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class Continue(OperationSearcherMixin):
    """
    Operation to skip the current operation ad move to the next one.
    Can be used in selection/operation pipelines with `IfElse` operation
    to conditionally skip the operation.

    Example
    -------
    >>> from soupsavvy.operations import Continue, IfElse, Operation
    ... operation = IfElse(
    ...     lambda x: x == 0,
    ...     Continue(),
    ...     Operation(lambda x: x / 100),
    ... ) | Operation(lambda x: x - 1)
    ... operation.execute(0)
    -1
    ... operation.execute(100)
    0

    If `Continue` operation is executed,
    operation is skipped and the next one is executed.
    """

    def _execute(self, arg: Any) -> Any:
        return arg

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Continue)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



"""
Module for operations wrappers,
operations that control behavior of the wrapped operation.

Classes
-------
- `SkipNone` - Skips the operation if the input is None.
- `Suppress` - Suppresses exceptions raised by the operation.
"""

from __future__ import annotations

from typing import Any, Type, Union

import soupsavvy.exceptions as exc
from soupsavvy.base import BaseOperation, OperationSearcherMixin, check_operation

# allowed exception types to suppress
ExceptionCategory = Union[Type[Exception], tuple[Type[Exception], ...]]


class OperationWrapper(OperationSearcherMixin):
    """
    A wrapper class for operations that extends the base operation functionality.
    Acts as a higher order operation, which controls behavior of the wrapped operation.
    """

    def __init__(self, operation: BaseOperation) -> None:
        """
        Initialize Operation Wrapper.

        Parameters
        ----------
        operation : BaseOperation
            The operation to be wrapped.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        self.operation = check_operation(operation)

    def __eq__(self, x: Any) -> bool:
        """
        Check if two `OperationWrapper` instances are equal.
        They need to be of the same class and wrap the same operation.

        Parameters
        ----------
        x : Any
            The object to compare with.

        Returns
        -------
        bool
            True if the instances are equal, False otherwise.
        """
        return isinstance(x, self.__class__) and self.operation == x.operation

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.operation})"

    def __repr__(self) -> str:
        return str(self)


class SkipNone(OperationWrapper):
    """
    A wrapper that skips the operation if the input is None.
    Used to prevent exceptions where it's safe and expected to skip the operation.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... from soupsavvy.models import SkipNone
    ... operation = SkipNone(Text())
    ... operation.execute(None)
    None

    When element was not found, which can be expected, skips operation
    and returns None.
    """

    def _execute(self, arg: Any) -> Any:
        if arg is None:
            return None

        return self.operation.execute(arg)


class Suppress(OperationWrapper):
    """
    A wrapper that executes the operation and suppresses exceptions raised,
    returning None instead. Used to catch exceptions where it's expected
    this might happen.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int))
    ... operation.execute("")
    None

    This can be used with `Default` operation to provide a default value
    when the operation fails.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Default(Suppress(Operation(int)), 0)
    ... operation.execute("")
    None

    Operations in example can be used to try to convert string to integer
    from text of tag, that can potentially be empty. In such case, if it's not required,
    default can be set to None or known value.

    Suppress also accepts category of exceptions to suppress, by default it suppresses
    all exceptions that inherit from Exception.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int), category=ValueError)
    ... operation.execute("not an integer")
    None

    Category can be a tuple of exceptions as well, `issubclass` is used to check
    if the cause of exception is subclass of provided category.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... from soupsavvy.models import Suppress
    ... operation = Suppress(Operation(int), category=(AttributeError, ValueError))
    ... operation.execute("not an integer")
    FailedOperationExecution
    """

    def __init__(
        self,
        operation: BaseOperation,
        category: ExceptionCategory = Exception,
    ) -> None:
        """
        Initialize `Suppress` operation instance.

        Parameters
        ----------
        operation : BaseOperation
            The operation to be wrapped.
        category : Type[Exception] | tuple[Type[Exception], ...], optional
            The exception type(s) to suppress. By default, suppresses all exceptions
            that inherit from `Exception`.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        super().__init__(operation)
        self._category = category

    def _execute(self, arg: Any) -> Any:
        try:
            return self.operation.execute(arg)
        except exc.FailedOperationExecution as e:
            cause = e.__cause__.__class__

            if not issubclass(cause, self._category):
                raise
            return None



"""Module with constants and defaults for model classes."""

# default recursive value for finding fields of model within the scope
DEFAULT_RECURSIVE = True
# default strict value for finding fields of model within the scope
DEFAULT_STRICT = False

# attribute to set on method that post-processes field value of model
POST_ATTR = "__post_process_method__"
# name of the attribute that keep tracks of initialization state
INITIALIZED = "_initialized"

SCOPE = "__scope__"
INHERIT_FIELDS = "__inherit_fields__"
POST_PROCESSORS = "__post_processors__"
FROZEN = "__frozen__"

SPECIAL_FIELDS = {SCOPE, INHERIT_FIELDS, POST_PROCESSORS, FROZEN}

# based model classes that skip initialization checks
BASE_MODELS = {"BaseModel"}



"""
Module with selectors that represent logical operations on multiple selectors.

Classes
-------
- `AndSelector` - intersection of multiple selectors (&)
- `NotSelector` - negation of a selector (~)
- `XORSelector` - exclusive OR of multiple selectors (^)
- `SelectorList` - counterpart of CSS selector list or :is() pseudo-class (|)
- `OrSelector` - union of multiple selectors - alias of `SelectorList` (|)
"""

from collections import Counter
from functools import reduce
from typing import Optional

from bs4 import Tag

from soupsavvy.base import CompositeSoupSelector, SoupSelector
from soupsavvy.utils.selector_utils import TagIterator, TagResultSet, UniqueTag


class SelectorList(CompositeSoupSelector):
    """
    Counterpart of CSS selector list.
    At least one selector from list must match the element to be included.

    Example
    -------
    >>> SelectorList(TypeSelector("a"), AttributeSelector(name="class", value="widget"))

    matches all elements that have "a" tag name OR 'class' attribute "widget".

    Example
    -------
    >>> <a>Hello World</a> ✔️
    >>> <div class="widget">Hello World</div> ✔️
    >>> <div>Hello Python</div> ❌

    Object can be created as well by using `pipe` operator `|` on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("a") | ClassSelector("widget")

    CSS counterpart can be represented as:

    Example
    -------
    >>> a, .widget
    >>> :is(a, .widget)

    Notes
    -----
    For more information on selector list, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Selector_list
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes `SelectorList` object with provided positional arguments.

        Parameters
        ----------
        selectors: SoupSelector
            At least two `SoupSelector` objects to match accepted as positional arguments.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of SoupSelector.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        results = TagResultSet()

        for selector in self.selectors:
            results |= TagResultSet(
                selector.find_all(tag, recursive=recursive),
            )

        # keep order of tags and limit
        results = TagResultSet(list(TagIterator(tag, recursive=recursive))) & results
        return results.fetch(limit)


# alias of `SelectorList`
OrSelector = SelectorList


class NotSelector(CompositeSoupSelector):
    """
    Selector for finding elements that do not match provided selector(s).
    Counterpart of CSS :not() pseudo-class.

    Example
    -------
    >>> NotSelector(TypeSelector("div"))

    matches all elements that do not have "div" tag name.

    Example
    -------
    >>> <span> class="widget">Hello World</span> ✔️
    >>> <div class="menu">Hello World</div> ❌

    Object can be created as well by using `bitwise NOT` operator `~`
    on `SoupSelector` object.

    Example
    -------
    >>> ~TypeSelector("div")

    Which is equivalent to the first example.

    This is CSS counterpart of :not() pseudo-class.

    Example
    -------
    >>> :not(div)

    Notes
    -----
    For more information on :not() pseudo-class, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:not
    """

    def __init__(
        self,
        selector: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes `NotSelectors` object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            At least one `SoupSelector` objects to negate match
            accepted as positional arguments.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of `SoupSelector`.
        """
        self._multiple = bool(selectors)
        super().__init__([selector, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        matching = reduce(
            TagResultSet.__or__,
            (
                TagResultSet(step.find_all(tag, recursive=recursive))
                for step in self.selectors
            ),
        )
        all_tags = list(TagIterator(tag, recursive=recursive))
        result = TagResultSet(all_tags) - matching
        return result.fetch(limit)

    def __invert__(self) -> SoupSelector:
        """
        Overrides __invert__ method to cancel out negation by returning
        the tag in case of single selector, or `SelectorList` in case of multiple.
        """
        if not self._multiple:
            return self.selectors[0]

        return SelectorList(*self.selectors)


class AndSelector(CompositeSoupSelector):
    """
    Selector representing an intersection of multiple selectors,
    where element must be matched by all provided selectors.
    Counterpart of CSS compound selector.

    Example
    -------
    >>> AndSelector(TypeSelector("div"), ClassSelector("widget"))

    matches all elements that have "div" tag name AND 'class' attribute "widget".

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <span class="widget">Hello World</span> ❌
    >>> <div class="menu">Hello World</div> ❌

    Object can be created as well by using `bitwise AND` operator `&`
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") & ClassSelector("widget")

    CSS counterpart can be represented as:

    Example
    -------
    >>> div.widget

    Notes
    -----
    For more information on compound selectors ,see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#compound_selector
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes `AndSelector` object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            At least two `SoupSelector` objects to match accepted as positional arguments.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of `SoupSelector`.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        matching = reduce(
            TagResultSet.__and__,
            (
                TagResultSet(step.find_all(tag, recursive=recursive))
                for step in self.selectors
            ),
        )
        return matching.fetch(limit)


class XORSelector(CompositeSoupSelector):
    """
    Selector representing an exclusive OR of multiple selectors,
    where element must be matched by exactly one of them.

    Example
    -------
    >>> XORSelector(TypeSelector("div"), ClassSelector("widget"))

    Matches all elements that have either "div" tag name or 'class' attribute "widget".
    Elements with both "div" tag name and 'class' attribute "widget" do not match selector.

    Object can be created as well by using `xor` operator `^` (caret)
    on `SoupSelector` objects.

    Example
    -------
    >>> TypeSelector("div") ^ ClassSelector("widget")

    This is a shortcut for defining XOR operation between two selectors like this:

    Example
    -------
    >>> selector1 = TypeSelector("div")
    ... selector2 = ClassSelector("widget")
    ... xor = (selector1 & (~selector2)) | ((~selector1) & selector2)
    """

    def __init__(
        self,
        selector1: SoupSelector,
        selector2: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes `XORSelector` object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            At least two `SoupSelector` objects to match accepted as positional arguments.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of `SoupSelector`.
        """
        super().__init__([selector1, selector2, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        unique = (
            UniqueTag(element)
            for step in self.selectors
            for element in step.find_all(tag, recursive=recursive)
        )
        results = TagResultSet(
            [element.tag for element, count in Counter(unique).items() if count == 1]
        )
        # keep order of tags and limit
        results = TagResultSet(list(TagIterator(tag, recursive=recursive))) & results
        return results.fetch(limit)



"""
This module is deprecated and will be removed in the future.
Use `soupsavvy.operators` module instead.
"""

import warnings

from soupsavvy.base import SoupSelector
from soupsavvy.selectors.logical import (
    AndSelector,
    NotSelector,
    SelectorList,
    XORSelector,
)
from soupsavvy.selectors.relative import HasSelector

warnings.warn(
    "soupsavvy.api module is deprecated, use soupsavvy.operators",
    DeprecationWarning,
    stacklevel=2,
)


def is_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> SelectorList:
    """
    Returns an union of multiple soup selectors.

    Parameters
    ----------
    selectors: SoupSelector
         At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    SelectorList
        `SelectorList` object that represents union of provided selectors.

    Example
    -------
    >>> is_(TypeSelector("h1"), TypeSelector("h2"))
    ... where(TypeSelector("h1"), TypeSelector("h2"))

    This is CSS counterpart of `selector list`, `:is()` or `:where` pseudo-classes.

    Example
    -------
    >>> h1, h1
    >>> :is(h1, h2)
    >>> :where(h1, h2)

    Aliases
    -------
    - `where` - to imitate css `:where` pseudo-class
    - `or_` - to imitate logical operator `or`

    Notes
    -----
    For more information on css counterparts, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:is
    - https://developer.mozilla.org/en-US/docs/Web/CSS/:where
    - https://developer.mozilla.org/en-US/docs/Web/CSS/Selector_list
    """
    return SelectorList(selector1, selector2, *selectors)


where = is_
or_ = is_


def not_(selector: SoupSelector, /, *selectors: SoupSelector) -> NotSelector:
    """
    Returns the negation of multiple soup selectors.

    Parameters
    ----------
    selectors: SoupSelector
        `SoupSelector` object(s) to negate match accepted as positional arguments.

    Returns
    -------
    NotSelector
        `NotSelector` object that represents negation of provided selector(s).

    Example
    -------
    >>> not_(TypeSelector("div"), TypeSelector("a"))


    This is the counterpart of CSS `:not()` negation pseudo-class,
    that represents elements that do not match a list of selectors.

    Example
    -------
    >>> :not(div, a)

    Notes
    -----
    For more information on `:not()` pseudo-class, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:not
    """
    return NotSelector(selector, *selectors)


def and_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> AndSelector:
    """
    Returns an intersection of multiple soup selectors.
    Element must match all of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    AndSelector
        `AndSelector` object that represents intersection of provided selectors.

    Example
    -------
    >>> and_(TypeSelector("div"), ClassSelector("widget"))

    This is the counterpart of CSS compound selector (selector concatenation):

    Example
    -------
    >>> div.widget

    Notes
    -----
    For more information on compound selectors, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#compound_selector
    """
    return AndSelector(selector1, selector2, *selectors)


def has(
    selector: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> HasSelector:
    """
    Returns `HasSelector`, whose behavior imitate css `:has()` pseudo-class,
    that represents elements if any of the relative selectors that are passed as an argument
    match at least one element when anchored against this element.

    Parameters
    ----------
    selectors: SoupSelector
        `SoupSelector` object(s) to match accepted as positional arguments.

    Returns
    -------
    HasSelector
        `HasSelector` object that represents `:has()` pseudo-class behavior.

    Example
    -------
    >>> has_(TypeSelector("div"), TypeSelector("a"))

    This is the counterpart of CSS `:has()` pseudo-class:

    Example
    -------
    >>> :has(div, a)

    Notes
    -----
    For more information on `:has()` pseudo-class, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:has
    """
    return HasSelector(selector, *selectors)


def xor(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> XORSelector:
    """
    Returns an symmetric difference of multiple soup selectors.
    Element must match exactly one of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    XORSelector
       `XORSelector` object that represents symmetric difference of provided selectors.

    Example
    -------
    >>> xor(TypeSelector("div"), ClassSelector("widget"))

    This is a shortcut for defining `XOR` operation between two selectors like this:

    Example
    -------
    >>> selector1 = TypeSelector("div")
    ... selector2 = ClassSelector("widget")
    ... xor = (selector1 & (~selector2)) | ((~selector1) & selector2)
    """
    return XORSelector(selector1, selector2, *selectors)



"""
Module with miscellaneous selectors.

Classes
-------
- `TypeSelector` - combines type and attribute selectors
- `PatternSelector` - matches elements based on text content and selector
- `UniversalSelector` - universal selector (*)
- `SelfSelector` - matches the element itself
- `ExpressionSelector` - matches elements based on user-defined function
"""

import itertools
from collections.abc import Callable
from dataclasses import dataclass
from typing import Optional, Pattern

from bs4 import SoupStrainer, Tag
from typing_extensions import deprecated

import soupsavvy.selectors.namespace as ns
from soupsavvy.base import SelectableCSS, SoupSelector
from soupsavvy.utils.selector_utils import TagIterator


@dataclass
class TypeSelector(SoupSelector, SelectableCSS):
    """
    Selector for finding elements based on tag name (type).
    Counterpart of css type selectors.

    Example
    -------
    >>> TypeSelector("div")

    matches all elements that have "div" tag name.

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <a href="/shop">Hello World</a> ❌

    CSS counterpart can be represented as:

    Example
    -------
    >>> div

    And can be retrieved with `css` property.

    Example
    -------
    >>> TypeSelector("div").css
    "div"

    Parameters
    ----------
    name : str
        Tag name of the element ex. "a", "div".

    Notes
    -----
    For more information about type selectors, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Type_selectors
    """

    name: str

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        params = {ns.NAME: self.name}
        return tag.find_all(**params, recursive=recursive, limit=limit)

    @property
    def css(self) -> str:
        # css selector for tag name is just the tag name ex. "div"
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeSelector):
            return False

        # TypeSelectors produce the same results if names of the tag are the same
        return self.name == other.name


@dataclass
class PatternSelector(SoupSelector):
    """
    Selector for finding elements based on text content pattern.

    Example
    -------
    >>> PatternSelector(pattern="Hello World")

    matches all element with exact text content "Hello World".

    Example
    -------
    >>> <div>Hello World</div> ✔️
    >>> <div>Hello Python</div> ❌
    >>> <div>Hello World 3</div> ❌

    In case of using regex pattern, `re.search` is used to match the attribute value.

    Example
    -------
    >>> PatternSelector(pattern=re.compile(r"[0-9]+"))

    matches all elements with text content containing at least one digit.

    Example
    -------
    >>> <div>Hello World 123</div> ✔️
    >>> <div>Hello World</div> ❌

    Parameters
    ----------
    pattern: str | Pattern
        Pattern to match text of the element. Can be a string for exact match
        or `Pattern` for any more complex regular expressions.

    Notes
    -----
    Due to `bs4` implementation, element does not match the pattern if it has any children.
    Only leaf nodes can be returned by `PatternSelector` find methods.
    """

    pattern: ns.PatternType

    def __post_init__(self) -> None:
        """Sets up compiled regex pattern used for `SoupStrainer` in find methods."""
        self._pattern = (
            str(self.pattern) if not isinstance(self.pattern, Pattern) else self.pattern
        )

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        iterator = TagIterator(tag, recursive=recursive)
        strainer = SoupStrainer(string=self._pattern)  # type: ignore
        filter_ = filter(strainer.search_tag, iterator)
        return list(itertools.islice(filter_, limit))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PatternSelector):
            return False

        return self._pattern == other._pattern


@dataclass
class UniversalSelector(SoupSelector, SelectableCSS):
    """
    Selector representing a wildcard pattern, that matches all elements in the html page.

    Example
    -------
    >>> UniversalSelector()

    CSS counterpart can be represented as:

    Example
    -------
    >>> *

    And can be retrieved with `css` property.

    Example
    -------
    >>> UniversalSelector().css
    "*"

    Notes
    -----
    For more information on universal selector, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Universal_selectors
    """

    def find_all(
        self, tag: Tag, recursive: bool = True, limit: Optional[int] = None
    ) -> list[Tag]:
        return tag.find_all(recursive=recursive, limit=limit)

    @property
    def css(self) -> str:
        """Returns wildcard css selector matching all elements in the markup."""
        return ns.CSS_SELECTOR_WILDCARD

    def __eq__(self, other: object) -> bool:
        # UniversalSelector is always the same
        return isinstance(other, UniversalSelector)


@deprecated(f"'AnyTagSelector' is deprecated, use 'UniversalSelector' class instead.")
class AnyTagSelector(UniversalSelector):
    """Alias for `UniversalSelector` class. Deprecated component."""


class SelfSelector(SoupSelector):
    """
    Selector matching only the element itself.
    Convenience component that can be used for compatibility.

    Example
    -------
    >>> SelfSelector()

    always matches the tag that is passed to the find methods.

    Notes
    -----
    Can be used in user-defined model for scope if element itself is the scope.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return [tag]

    def __eq__(self, other: object) -> bool:
        # needs to be the same type
        return isinstance(other, SelfSelector)


@dataclass
class ExpressionSelector(SoupSelector):
    """
    Selector that matches elements based on a user-defined function (predicate),
    that is used as filter for `bs4` object.

    Applies predicate to each element and returns those that satisfy the condition.

    Parameters
    ----------
    f : Callable[[Tag], bool]
        A user-defined function (predicate) that determines whether
        the element should be selected.

    Examples
    --------
    >>> selector = ExpressionSelector(lambda x: x.name not in {"a", "div"})
    ... selector.find(soup)

    Notes
    -----
    Any exceptions should be handled inside provided function.
    If raised, it will be propagated to the caller.
    """

    f: Callable[[Tag], bool]

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        iterator = TagIterator(tag, recursive=recursive)
        filter_ = filter(self.f, iterator)
        return list(itertools.islice(filter_, limit))

    def __eq__(self, other) -> bool:
        if not isinstance(other, ExpressionSelector):
            return False

        return self.f is other.f



"""
Module with selectors that search for elements based on their attributes.

Classes
-------
- `AttributeSelector` - Selects element based on any attribute value.
- `IdSelector` - Selects element based on 'id' attribute value.
- `ClassSelector` - Selects element based on 'class' attribute value.
"""

import re
from dataclasses import dataclass
from typing import Optional, Pattern

from bs4 import Tag

import soupsavvy.selectors.namespace as ns
from soupsavvy.base import SoupSelector
from soupsavvy.selectors.namespace import PatternType


@dataclass
class AttributeSelector(SoupSelector):
    """
    Selector for searching element based on its attribute value.
    Counterpart of css attribute selectors, that extends its capability
    with regex pattern matching.

    Example
    -------
    >>> AttributeSelector(name="role", value="widget")

    matches all elements that have 'role' attribute with value "widget".

    Example
    -------
    >>> <div role="widget">Hello World</div> ✔️
    >>> <div class="menu">Hello World</div> ❌
    >>> <div role="menu">Hello World</div> ❌

    CSS counterpart can be represented as:

    Example
    -------
    >>> [role="widget"]

    In case of using regex pattern, `re.search` is used to match the attribute value.

    Example
    -------
    >>> AttributeSelector(name="href", value=re.compile(r"wikipedia"))

    Parameters
    ----------
    name : str
        HTML element attribute name ex. "class", "href"
    value : str | Pattern, optional
        Value of the attribute to match.
        By default None, if not provided, default pattern matching any sequence
        of characters is used.

    Notes
    -----
    For more information about attribute selectors, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Attribute_selectors
    """

    name: str
    value: Optional[PatternType] = None

    def __post_init__(self) -> None:
        """Sets pattern attribute used in `SoupSelector` find operations."""
        self._pattern = self._parse_pattern()

    def _parse_pattern(self) -> PatternType:
        """Parses pattern used in find methods based on provided init parameters."""
        # if value was not provided, fall back to default pattern
        if self.value is None:
            return re.compile(ns.DEFAULT_PATTERN)
        # cast value to string if not a regex pattern
        if not isinstance(self.value, Pattern):
            return str(self.value)
        # value is already a compiled regex pattern
        return self.value

    def find_all(
        self,
        tag: ns.Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        params = {ns.ATTRS: {self.name: self._pattern}}
        return tag.find_all(**params, recursive=recursive, limit=limit)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AttributeSelector):
            return False

        # pattern is what is used in find methods
        return self._pattern == other._pattern and self.name == other.name


class SpecificAttributeSelector(AttributeSelector):
    """
    Base class for specific attribute selectors,
    that wraps `AttributeSelector` with default attribute name for user convenience.

    Child classes should define _NAME attribute with default attribute name,
    that will be used in the `AttributeSelector`.
    """

    _NAME: str

    def __init__(self, value: Optional[PatternType] = None) -> None:
        """
        Initializes specific attribute selector with default attribute name.

        Parameters
        ----------
        value : str | Pattern, optional
            Value of the attribute to match.
            By default None, if not provided, default pattern matching any sequence
            of characters is used.
        """
        super().__init__(name=self._NAME, value=value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value!r})"


class IdSelector(SpecificAttributeSelector):
    """
    Specific `AttributeSelector` for matching elements based on 'id' attribute value.

    Example
    -------
    >>> IdSelector("main")

    matches all elements that have 'id' attribute with value "main".

    Example
    -------
    >>> <div id="main">Hello World</div> ✔️
    >>> <div id="content">Hello World</div> ❌

    `IdSelector` is a convenience wrapper for `AttributeSelector`,
    thus example above is equivalent to using:

    >>> AttributeSelector(name="id", value="main")

    CSS counterpart can be represented as:

    Example
    -------
    >>> #main

    In case of using regex pattern, `re.search` is used to match the attribute value.

    Example
    -------
    >>> IdSelector(re.compile(r"content[0-9]+"))

    Notes
    -----
    For more information about id attribute, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/ID_selectors
    """

    _NAME = "id"


class ClassSelector(SpecificAttributeSelector):
    """
    Specific `AttributeSelector` for matching elements based on 'class' attribute value.

    Example
    -------
    >>> ClassSelector("widget")

    matches all elements that have 'class' attribute with value "widget".

    Example
    -------
    >>> <div class="widget">Hello World</div> ✔️
    >>> <div class="content">Hello World</div> ❌

    `ClassSelector` is a convenience wrapper for `AttributeSelector`,
    thus example above is equivalent to using:

    >>> AttributeSelector(name="class", value="widget")

    CSS counterpart can be represented as:

    Example
    -------
    >>> .widget

    In case of using regex pattern, `re.search` is used to match the attribute value.

    Example
    -------
    >>> ClassSelector(re.compile(r"nav"))

    Notes
    -----
    For more information about class attribute, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/Class_selectors
    """

    _NAME = "class"



"""
Module with relative selectors and utility components.
They are used for selecting elements based on their relation to the anchor element.

Classes
-------
- `RelativeChild` - matches direct children of the anchor element
- `RelativeDescendant` - matches descendants of the anchor element
- `RelativeNextSibling` - matches next sibling of the anchor element
- `RelativeSubsequentSibling` - matches subsequent siblings of the anchor element
- `RelativeParent` - matches parent of the anchor element
- `RelativeAncestor` - matches ancestors of the anchor element
- `HasSelector` - selects elements based on matching reference elements
- `Anchor` - Anchor object for easily creating relative selectors
"""

from collections.abc import Callable
from typing import Optional

from bs4 import Tag

from soupsavvy.base import CompositeSoupSelector, SoupSelector, check_selector
from soupsavvy.utils.selector_utils import TagIterator, TagResultSet


class RelativeSelector(SoupSelector):
    """
    Base class for relative selectors, that are used to find tags relative
    to the tag that is being searched, which is considered an anchor.

    CSS definition of relative selectors state, that it is a selector representing
    an element relative to one or more anchor elements preceded by a combinator.

    In this use case, the anchor element is the tag that is being searched, and
    the combinator is the logic of specific relative selector that is used.

    Example
    -------
    >>> selector = Anchor > TypeSelector("div")
    ... selector.find_all(tag)

    Uses `RelativeChild` selector to find any `div` tag that is a direct child of the
    tag that is being searched (passed as an argument).

    In css such selectors can be used for example in `:has` pseudo-class, where selector
    is anchored to the element:

    Example
    -------
    >>> :has(> div)

    Selects any element that has a direct child `div` tag.

    Notes
    -------
    Recursive parameter is ignored in relative selectors,
    as they have their own logic of searching the document.
    """

    def __init__(self, selector: SoupSelector) -> None:
        """
        Initializes RelativeSelector instance with specified selector.

        Parameters
        ----------
        selector : SoupSelector
            Selector that is used to find tags relative to the anchor tag.
        """
        self._selector = check_selector(selector)

    @property
    def selector(self) -> SoupSelector:
        """
        Returns selector used to find elements relative to the anchor element
        in this relative selector.

        Returns
        -------
        SoupSelector
            Selector used for searching elements relative to the anchor element.
        """
        return self._selector

    def __eq__(self, other: object) -> bool:
        # check for RelativeSelector type for type checking sake
        if not isinstance(other, RelativeSelector):
            return False
        elif type(self) is not type(other):
            # checking for exact type match - isinstance(other, self.__class__)
            # when other is subclass of self.__class__ would call other.__eq__(self)
            # which is not desired behavior, as it returns False
            return False

        return self.selector == other.selector

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector})"

    def __repr__(self) -> str:
        return str(self)


class BaseRelativeSibling(RelativeSelector):
    """
    Base class with implementation for relative sibling selectors,
    searches for next sibling(s) of the anchor tag.
    Child class needs to define:
    - '_limit' - class attribute to specify how many next siblings to search for.
    - '_func' - class attribute to specify which method to use for finding siblings.
    """

    # limit of next siblings to check
    _limit: Optional[int]
    # function to use for finding siblings
    _func: Callable[..., list]

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        search = tag.find_parents()[0]
        # find all sibling tags that match the selector
        matching = TagResultSet(
            self.selector.find_all(search, recursive=False),
        )
        siblings = TagResultSet(self.__class__._func(tag, limit=self._limit))
        # find intersection between two sets
        matches = matching & siblings
        return matches.fetch(limit)


class BaseAncestorSelector(RelativeSelector):
    """
    Base class with implementation for ancestor selectors,
    searches for ancestor(s) of the anchor tag.
    Child class needs to define:
    - '_limit' - class attribute to specify how many ancestors to search for.
    """

    _limit: Optional[int]

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        limit = limit or self._limit
        # get max number of ancestors that can possibly be returned
        ancestors = tag.find_parents(limit=self._limit)

        if not ancestors:
            # if no ancestors, make no sense to search
            return []

        search = ancestors[-1].parent or ancestors[-1]
        # search within parent of last ancestor
        matching = TagResultSet(self.selector.find_all(search))
        matches = TagResultSet(ancestors) & matching
        return matches.fetch(limit)


class RelativeChild(RelativeSelector):
    """
    Selector for finding direct children of the anchor element.

    Example
    -------
    >>> RelativeChild(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ❌
    >>> <div><a></a></div> ❌

    It can be created with `Anchor` instance as well with use of `gt` operator `>`:

    Example
    -------
    >>> Anchor > TypeSelector("p")

    Notes
    -------
    Behavior of `RelativeChild` selector is equivalent to using find methods of
    selector with `recursive=False` and is implemented to support 'HasSelector'
    and `ChildCombinator` selectors.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return self.selector.find_all(tag, recursive=False, limit=limit)


class RelativeDescendant(RelativeSelector):
    """
    Selector for finding descendants of the anchor element.

    Example
    -------
    >>> RelativeDescendant(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ✔️
    >>> <div><a></a></div> ❌
    >>> <div></div><p></p> ❌

    It can be created with `Anchor` instance as well with use of `right shift` operator `>>`:

    Example
    -------
    >>> Anchor >> TypeSelector("p")

    Notes
    -------
    Behavior of `RelativeDescendant` selector is equivalent to using find methods of
    selector with default `recursive=True` and is implemented to support
    'HasSelector' and `DescendantCombinator` selectors.
    """

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return self.selector.find_all(tag, recursive=True, limit=limit)


class RelativeNextSibling(BaseRelativeSibling):
    """
    Selector for finding next sibling of the anchor element.

    Example
    -------
    >>> RelativeNextSibling(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div></div><p></p> ✔️
    >>> <div></div><a></a><p></p>  ❌
    >>> <p></p><div></div> ❌

    It can be created with `Anchor` instance as well with use of `plus` operator `+`:

    Example
    -------
    >>> Anchor + TypeSelector("p")
    """

    _limit = 1
    _func = Tag.find_next_siblings


class RelativeSubsequentSibling(BaseRelativeSibling):
    """
    Selector for finding subsequent siblings of the anchor element.

    Example
    -------
    >>> RelativeSubsequentSibling(TypeSelector("p"))

    when 'div' element is passed into find methods:

    Example
    -------
    >>> <div></div><p></p> ✔️
    >>> <div></div><a></a><p></p> ✔️
    >>> <p></p><div></div> ❌
    >>> <div></div><span><p></p></span> ❌

    It can be created with `Anchor` instance as well
    with use of `multiplication` operator `*`:

    Example
    -------
    >>> Anchor * TypeSelector("p")
    """

    _limit = None
    _func = Tag.find_next_siblings


class RelativeParent(BaseAncestorSelector):
    """
    Selector for finding parent of the anchor element.

    Example
    -------
    >>> RelativeParent(TypeSelector("div"))

    when 'p' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ❌
    >>> <span><p></p></span> ❌

    Although this combinator does not have its counterpart in CSS, it can be
    represented as has selector, where child combinator is explicitly stated:

    Example
    -------
    >>> div:has(> p)

    It can be created with `Anchor` instance as well with use of `lt` operator `<`:

    Example
    -------
    >>> Anchor < TypeSelector("div")

    Notes
    -------
    `RelativeParent` selector ignores `recursive` parameter,
    as it is always searches only for parent of the anchor element,
    `find_all` method can return at most one element (parent).
    """

    _limit = 1


class RelativeAncestor(BaseAncestorSelector):
    """
    Selector for finding ancestors of the anchor element.

    Example
    -------
    >>> RelativeAncestor(TypeSelector("div"))

    when 'p' element is passed into find methods:

    Example
    -------
    >>> <div><p></p></div> ✔️
    >>> <div><a><p></p></a></div> ✔️
    >>> <span><p></p></span> ❌
    >>> <p></p><div></div> ❌


    Although this combinator does not have its counterpart in CSS, it can be
    represented as has selector, where descendant combinator is implied:

    Example
    -------
    >>> div:has(p)

    It can be created with `Anchor` instance as well with use of `left shift` operator `<<`:

    Example
    -------
    >>> Anchor << TypeSelector("div")

    Notes
    -------
    `RelativeAncestor` selector ignores `recursive` parameter,
    as it is always searches among all ancestors of the anchor element.
    """

    _limit = None


class Anchor_:
    """
    Shortcut component used to create relative selectors
    in a more readable way with use of operators.

    `Anchor_` is an internal class and can be considered a singleton.
    It's advisable to use the `Anchor` instance instead of creating new objects.

    `Anchor` supports the following operators:
    - `>`: `RelativeChild`

    Example
    -------
    >>> Anchor > TypeSelector("div")

    Creates `RelativeChild` selector, that selects any div tag that is a direct
    child of the tag that is being searched.

    - `>>`: `RelativeDescendant`

    Example
    -------
    >>> Anchor >> TypeSelector("div")

    Creates `RelativeDescendant` selector, that selects any div tag that is a descendant
    of the tag that is being searched. This is default behavior of selectors,
    and is equivalent to using the `TypeSelector` directly, but is implemented
    for the sake of consistency.

    - `+`: `RelativeNextSibling`

    Example
    -------
    >>> Anchor + TypeSelector("div")

    Creates `RelativeNextSibling` selector, that selects any div tag that is next
    sibling of the tag that is being searched, it can logically return at most one tag.

    - `*`: `RelativeSubsequentSibling`

    Example
    -------
    >>> Anchor * TypeSelector("div")

    Creates `RelativeSubsequentSibling` selector, that selects any div tag
    that is a subsequent sibling of the tag that is being searched.

    - `<`: `RelativeParent`

    Example
    -------
    >>> Anchor < TypeSelector("div")

    Creates `RelativeParent` selector, that selects any div tag
    that is a parent of the tag that is being searched.

    - `<<`: `RelativeAncestor`

    Example
    -------
    >>> Anchor << TypeSelector("div")

    Creates `RelativeAncestor` selector, that selects any div tag
    that is an ancestor of the tag that is being searched.

    This imitates css selector relative selectors that are used for example in
    `:has` pseudo-class, that accepts relative selector list as an argument.

    Example
    -------
    >>> :has(> div, + a)

    This translated to `soupsavvy` would be:

    Example
    -------
    >>> HasSelector(Anchor > TypeSelector("div"), Anchor + TypeSelector("a"))

    Which would match any tag that has a direct child 'div' and a next sibling 'a' tag.
    Selected tag is the anchor tag that is being searched.

    Notes
    -------
    For more information on relative selectors, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#relative_selector
    """

    def __gt__(self, x: SoupSelector) -> RelativeChild:
        return RelativeChild(check_selector(x))

    def __rshift__(self, x: SoupSelector) -> RelativeDescendant:
        return RelativeDescendant(check_selector(x))

    def __lt__(self, x: SoupSelector) -> RelativeParent:
        return RelativeParent(check_selector(x))

    def __lshift__(self, x: SoupSelector) -> RelativeAncestor:
        return RelativeAncestor(check_selector(x))

    def __add__(self, x: SoupSelector) -> RelativeSelector:
        return RelativeNextSibling(check_selector(x))

    def __mul__(self, x: SoupSelector) -> RelativeSelector:
        return RelativeSubsequentSibling(check_selector(x))


# instance of Anchor class
Anchor = Anchor_()


class HasSelector(CompositeSoupSelector):
    """
    Selector for finding elements based on matching reference elements.

    Example
    -------
    >>> HasSelector(TypeSelector("div"))

    matches all elements that have any descendant with "div" tag name.
    It uses default combinator of relative selector, which is descendant combinator.

    Example
    -------
    >>> <span><div>Hello World</div></span> ✔️
    >>> <span><a>Hello World</a></span> ❌

    Other relative selectors can be used with `Anchor` element.

    Example
    -------
    ... HasSelector(Anchor > TypeSelector("div"))
    ... HasSelector(Anchor + TypeSelector("div"))

    or by using `RelativeSelector` components directly:

    Example
    -------
    ... HasSelector(RelativeChild(TypeSelector("div")))
    ... HasSelector(RelativeNextSibling(TypeSelector("div"))

    Example
    -------
    >>> <span><div>Hello World</div></span> ✔️
    >>> <span><a><div>Hello World</div></a></span> ❌

    In this case, HasSelector is anchored against any element, and matches only elements
    that have "div" tag name as a child.

    This is an equivalent of CSS :has() pseudo-class,
    that matches element if any of the relative selectors that are passed as an argument
    match element when anchored against it.

    Example
    -------
    >>> :has(div, a)
    >>> :has(+ div, > a)

    These examples translated to `soupsavvy` would be:

    Example
    -------
    ... HasSelector(TypeSelector("div"), TypeSelector("a"))
    ... HasSelector(Anchor + TypeSelector("div"), Anchor > TypeSelector("a"))

    Notes
    -----
    Passing `RelativeDescendant` selector into HasSelector is equivalent to using
    its selector directly, as descendant combinator is a default option.

    Example
    -------
    >>> HasSelector(RelativeDescendant(TypeSelector("div")))
    ... HasSelector(Anchor > TypeSelector("div"))
    ... HasSelector(TypeSelector("div"))

    Three of the above examples are equivalent.

    For more information on :has() pseudo-class, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:has
    """

    def __init__(
        self,
        selector: SoupSelector,
        /,
        *selectors: SoupSelector,
    ) -> None:
        """
        Initializes `HasSelector` object with provided positional arguments as selectors.

        Parameters
        ----------
        selectors: SoupSelector
            `SoupSelector` objects to match accepted as positional arguments.
            At least one selector is required to create `HasSelector`.

        Raises
        ------
        NotSoupSelectorException
            If any of provided parameters is not an instance of `SoupSelector`.
        """
        super().__init__([selector, *selectors])

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:

        elements = TagIterator(tag, recursive=recursive)
        matching: list[Tag] = []

        for element in elements:
            # we only care if anything matching was found
            if any(step.find(element) for step in self.selectors):
                matching.append(element)

                if len(matching) == limit:
                    break

        return matching



"""
Module with components used in nth selectors implementation.
Provide the ways of parsing and generating numbers from nth selectors.

Classes
-------
- `NthGenerator` - Generates a list of integers that match the nth CSS selector.
- `parse_nth` - Parses CSS nth selector and returns NthGenerator instance.
"""

import math
import re
from collections.abc import Callable, Generator
from dataclasses import dataclass
from itertools import count

# matches nth-selector that contains only step parameter - e.g. "2"
OFFSET_PATTERN = re.compile(r"^\d+$")
# matches nth-selector that contains only offset parameter - e.g. "-3n", "n"
STEP_PATTERN = re.compile(r"^-?\d*n$")
# matches nth-selector that contains both step and offset parameters - e.g. "-3n+2"
COMBINED_PATTERN = re.compile(r"^-?\d*n\+\d+$")


@dataclass
class NthGenerator:
    """
    Class for generating a list of integers that match the nth-child CSS selector.

    Parameters
    ----------
    step : int
        The step parameter of the nth-child CSS selector. Must be an integer.
    offset : int
        The offset parameter of the nth-child CSS selector.
        Must be a non-negative integer.

    Examples
    --------
    >>> nth = NthGenerator(step=2, offset=1)
    ... list(nth.generate(10))
    [1, 3, 5, 7, 9]

    >>> nth = NthGenerator(step=3, offset=2)
    ... list(nth.generate(10))
    [2, 5, 8]

    `generate` method has `stop` parameter that limits the range of integers generated.
    When having list of 10 sibling elements, the `stop` parameter can be set to 10
    to determine which elements match provided nth-child selector.

    Examples
    --------
    >>> nth = NthGenerator(step=-1, offset=3)
    ... elements = ["a", "b", "c", "d", "e", "f"]
    ... index = [x-1 for x in nth.generate(len(elements))]
    ... [elements[i] for i in index]
    ["d", "e", "f"]

    This selector matches first three children, and can be used the way shown above
    to filter out the elements that match the selector.
    """

    step: int
    offset: int

    def __post_init__(self):
        """
        Checks if passed parameters are valid.

        Raises
        ------
        TypeError
            If step or offset are not integers.
        ValueError
            If offset is negative.
        """
        if not isinstance(self.step, int):
            raise TypeError(
                f"Step parameter must be an integer, not {type(self.step)}."
            )
        if not isinstance(self.offset, int):
            raise TypeError(
                f"Offset parameter must be an integer, not {type(self.offset)}."
            )
        if self.offset < 0:
            raise ValueError("Offset must not be negative.")

    @property
    def _increasing(self) -> bool:
        """
        Returns true if linear function: y = ax + b where a = step and b = offset
        is increasing (step is positive), false otherwise.
        """
        return self.step > 0

    @property
    def _func(self) -> Callable[[int], int]:
        """
        Returns a linear function that represents the nth-child selector.
        y = ax + b where a = step and b = offset, which is used to generate
        the list of integers that match the selector.
        """
        a, b = self.step, self.offset
        return lambda x: a * x + b

    def generate(self, stop: int) -> Generator[int, None, None]:
        """
        Generates a list of integers that match the nth-child selector.

        Parameters
        ----------
        stop : int
            The stop parameter that limits the range of integers generated.
            Must be a positive integer.

        Yields
        ------
        int
            Integers that match the nth-child selector.
            Generated integers fall into the range from 1 to stop.

        Raises
        ------
        TypeError
            If stop parameter is not an integer.

        Examples
        --------
        >>> nth = NthGenerator(step=2, offset=1)
        >>> list(nth.generate(10))
        [1, 3, 5, 7, 9]

        >>> nth = NthGenerator(step=-3, offset=20)
        >>> list(nth.generate(8))
        [8, 5, 2]
        """
        if not isinstance(stop, int):
            raise TypeError(f"Stop parameter must be an integer, not {type(stop)}.")

        if stop < 1:
            return

        a, b = self.step, self.offset

        if a == 0:
            if 1 <= b <= stop:
                yield b
            return

        c = ((1 - b) / a) if self._increasing else ((stop - b) / a)
        start = max(math.ceil(c), 0)

        for x in count(start):  # pragma: no branch
            y = self._func(x)

            if not (1 <= y <= stop):
                break

            yield y


def parse_nth(selector: str) -> NthGenerator:
    """
    Parses CSS nth formula and return `NthGenerator` instance.

    Parameters
    ----------
    selector : str
        CSS nth selector string. Accepts all valid css nth formulas.

    Returns
    -------
    NthGenerator
        `NthGenerator` instance that represents the parsed nth selector
        and can be used for generating matching numbers.

    Raises
    ------
    ValueError
        If the selector is invalid.

    Handles "odd" and "even" literals and any number of whitespace characters
    in the selector string.

    Examples
    --------
    >>> nth = parse_nth("2")
    NthGenerator(step=0, offset=2)

    >>> nth = parse_nth("-3n+2")
    NthGenerator(step=-3, offset=2)

    >>> nth = parse_nth("odd")
    NthGenerator(step=2, offset=1)

    >>> nth = parse_nth("15n")
    NthGenerator(step=15, offset=0)
    """
    # handle whitespaces
    selector = selector.replace(" ", "")

    # handle special cases
    if selector == "odd":
        return NthGenerator(2, 1)
    if selector == "even":
        return NthGenerator(2, 0)

    # handle non-step cases
    if OFFSET_PATTERN.match(selector):
        return NthGenerator(0, int(selector))

    pattern_match = STEP_PATTERN.match(selector) or COMBINED_PATTERN.match(selector)

    if not pattern_match:
        raise ValueError(f"Invalid CSS nth selector: {selector}")

    # replaces n with no preceding digit with 1n - e.g. "n" -> "1n", "-n" -> "-1n"
    selector = re.sub(r"(\D|^)n", r"\g<1>1n", selector)
    # replaces missing offset with 0 - e.g. "2n" -> "2n0"
    selector = re.sub(r"n$", "n0", selector)

    step, offset = map(int, selector.split("n"))
    return NthGenerator(step, offset)



"""
Module with base model schema and its configuration.

Classes
-------
- `BaseModel` - Base class for defining search schemas.
- `Field` - Field configuration for search schema.
- `MigrationSchema` - Configuration of model migration.
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field as datafield
from functools import reduce
from typing import Any, Literal, Optional, Type, TypeVar, Union, overload

from bs4 import Tag
from typing_extensions import Self

import soupsavvy.exceptions as exc
import soupsavvy.models.constants as c
from soupsavvy.base import SoupSelector, check_selector
from soupsavvy.interfaces import Comparable, TagSearcher, TagSearcherExceptions

# Generic type variable for model migration
T = TypeVar("T")


@dataclass
class MigrationSchema:
    """
    Defines schema for model migration in need of providing additional parameters
    to the target model initialization in nested models.

    Attributes
    ----------
    target : Type
        Target model class to migrate to.
    params : dict, optional
        Additional parameters to pass to the target model initialization.
    """

    target: Type
    params: dict = datafield(default_factory=dict)


def post(field: str) -> Callable[[Callable], Callable]:
    """
    Decorator to mark a method as a post-processor for a model field.
    The method will be called after the field is extracted from the tag
    in model instance initialization.

    Example
    -------
    >>> class MyModel(BaseModel):
    ...    ...
    ...    field = ...
    ...
    ...    @post("field")
    ...    def post_process_field(self, value):
    ...        return value.strip()

    Methods of custom model class, that are decorated with `@post` decorator,
    must accept only one argument, which is the value of the field to be processed.
    """

    def decorator(func):
        setattr(func, c.POST_ATTR, field)
        return func

    return decorator


@dataclass
class Field(TagSearcher, Comparable):
    """
    Model field wrapper, that defined field metadata.
    Used for overwriting default behavior of attribute corresponding to field
    in model instance, similarly to dataclass field function.

    Attributes
    ----------
    selector : TagSearcher | type[BaseModel]
        Any searcher used in model as field.
    repr : bool, optional
        Whether the field should be included in the model's representation.
        Default is True.
    compare : bool, optional
        Whether the field should be included in the model's equality comparison
        as well as in the hash calculation. Default is True.
    migrate : bool, optional
        Whether the field should be migrated to the target model in model migration.
        Default is True.

    Example
    -------
    >>> class MyModel(BaseModel):
    ...    __scope__ = TypeSelector("p")
    ...
    ...    price = Text() | Operation(int)
    ...    element = Field(SelfSelector(), repr=False, compare=False, migrate=False)

    In this example, only price is relevant for model object. Element itself is just
    for reference and should not be included in model representation,
    comparison or migration.

    Using `Field` wrapper without any additional arguments
    is equivalent to default behavior.
    """

    selector: Union[TagSearcher, type[BaseModel]]
    repr: bool = True
    compare: bool = True
    migrate: bool = True

    def find_all(
        self, tag: Tag, recursive: bool = True, limit: Optional[int] = None
    ) -> list[Any]:
        return self.selector.find_all(tag, recursive=recursive, limit=limit)

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True) -> Any:
        return self.selector.find(tag, strict=strict, recursive=recursive)

    def __eq__(self, x: Any) -> bool:
        if not isinstance(x, Field):
            return False

        return all(
            [
                self.selector == x.selector,
                self.repr == x.repr,
                self.compare == x.compare,
                self.migrate == x.migrate,
            ]
        )

    def __repr__(self) -> str:
        return (
            f"Field({self.selector}, repr={self.repr}, compare={self.compare}, "
            f"migrate={self.migrate})"
        )


class ModelMeta(type(ABC)):
    """
    Metaclass for all models derived from `BaseModel`. This metaclass ensures that
    certain attributes and methods are defined
    and properly configured in each model class.

    It handles validation of provided attributes and controls inheritance of fields.
    Meta inherits from `type(ABC)` to avoid metaclass conflicts.

    Attributes
    ----------
    scope : SoupSelector
        Returns the `__scope__` attribute, which defines the scope
        within which the model is to be found.
    fields : dict[str, TagSearcher]
        Fields that defines the model and search operations.
    """

    def __init__(cls, name, bases, class_dict):
        """
        Initializes the model class and validates its attributes.
        For each user-defined model, which is a subclass of `BaseModel`,
        checks if scope and fields are properly defined.

        Raises
        ------
        ScopeNotDefinedException
            If the scope attribute is missing in the model class.
        FieldsNotDefinedException
            If no fields are defined in the model class.
        """
        super().__init__(name, bases, class_dict)

        setattr(cls, c.POST_PROCESSORS, {})
        post_processors = getattr(cls, c.POST_PROCESSORS)

        # Inherit post-processors from base classes
        for base in bases:
            post_processors.update(getattr(base, c.POST_PROCESSORS, {}))

        if name in c.BASE_MODELS:
            return

        scope = getattr(cls, c.SCOPE)

        if scope is None:
            raise exc.ScopeNotDefinedException(
                f"Missing '{c.SCOPE}' class attribute in '{name}'."
            )

        message = (
            f"'{c.SCOPE}' attribute of the model class '{cls.__name__}' "
            f"needs to be '{SoupSelector.__name__}' instance, got '{type(scope)}'."
        )
        check_selector(scope, message=message)

        fields = cls._get_fields()

        if not fields:
            raise exc.FieldsNotDefinedException(
                f"Model '{cls.__name__}' has no fields defined. At least one required."
            )

        # register post processors
        for name in dir(cls):
            obj = getattr(cls, name)

            if not callable(obj) or name.startswith("__"):
                continue

            field = getattr(obj, c.POST_ATTR, None)

            if field in fields:
                post_processors[field] = obj

    @property
    def scope(cls) -> SoupSelector:
        """
        Returns the `__scope__` attribute,
        which defines the scope selector for the model.

        Returns
        -------
        SoupSelector
            The scope selector used to identify the element in which
            the model is searched.
        """
        return getattr(cls, c.SCOPE)

    @property
    def fields(cls) -> dict[str, Field]:
        """
        Returns the fields of the model class with their
        respective TagSearcher instances. The fields are aggregated based on
        the `__inherit_fields__` setting.

        Returns
        -------
        dict[str, Field]
            A dictionary mapping field names to their respective `Field` instances.
        """
        return cls._get_fields()

    def _get_fields(cls) -> dict[str, Field]:
        """
        Returns the fields of the model with their respective `TagSearcher` instances
        based on the `__inherit_fields__` setting.

        If searcher is already provided in form of Field object, it is used directly.
        Otherwise, it is wrapped in `Field` object with default settings.
        """
        classes = (
            [
                base
                for base in reversed(cls.__mro__)
                if issubclass(base, BaseModel) and base is not BaseModel
            ]
            if getattr(cls, c.INHERIT_FIELDS)
            else [cls]
        )

        return reduce(
            dict.__or__,
            [
                {
                    key: value if isinstance(value, Field) else Field(value)
                    for key, value in class_.__dict__.items()
                    if (
                        # accepted field must be TagSearcher
                        isinstance(value, TagSearcher)
                        # or BaseModel subclass
                        or (isinstance(value, type) and issubclass(value, BaseModel))
                    )
                    and key not in c.SPECIAL_FIELDS
                }
                for class_ in classes
            ],
        )


class BaseModel(TagSearcher, Comparable, metaclass=ModelMeta):
    """Base class for all user-defined models in `soupsavvy`."""

    __scope__: SoupSelector = None  # type: ignore
    __inherit_fields__: bool = True
    __frozen__: bool = False

    __post_processors__: dict[str, Callable] = {}

    def __init__(self, **kwargs) -> None:
        """
        Initializes a model instance with provided field values.
        Model should not be initialized directly, but through the `find` methods.

        Parameters
        ----------
        kwargs : Any
            Field values to initialize the model with provided
            as keyword arguments.

        Raises
        ------
        MissingFieldsException
            If any required fields are missing from the provided kwargs.
        UnknownModelFieldException
            If any unknown fields are provided in the kwargs.
        """
        setattr(self, c.INITIALIZED, False)

        fields = self.__class__.fields.keys()

        for field in fields:
            try:
                value = kwargs.pop(field)
            except KeyError:
                raise exc.MissingFieldsException(
                    f"Cannot initialize model '{self.__class__.__name__}' "
                    f"without '{field}' field."
                )

            func = getattr(self, c.POST_PROCESSORS).get(field)

            if func is not None:
                value = func(self, value)

            setattr(self, field, value)

        if kwargs:
            raise exc.UnknownModelFieldException(
                f"Model '{self.__class__.__name__}' has no fields defined "
                f"for: {kwargs.keys()}"
            )

        self.__post_init__()
        setattr(self, c.INITIALIZED, True)

    def __post_init__(self) -> None:
        """
        Post-initialization hook. Can be overridden by subclasses to perform additional
        initialization steps after the base initialization.
        """

    @property
    def attributes(self) -> dict[str, Any]:
        """
        Returns a dictionary of model instance attributes representing
        model fields and their respective values.

        Returns
        -------
        dict[str, Any]
            A dictionary mapping model field names to their respective values.
        """
        return {field: getattr(self, field) for field in self.__class__.fields.keys()}

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: Literal[True] = ...,
        recursive: bool = ...,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: Literal[False] = ...,
        recursive: bool = ...,
    ) -> Optional[Self]: ...

    @overload
    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: bool = ...,
        recursive: bool = ...,
    ) -> Optional[Self]: ...

    @classmethod
    def find(
        cls,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Optional[Self]:
        """
        Searches for and returns an instance of the model within the provided tag.
        By default, perform recursive, non-strict search for model fields
        within the scope element.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to search within for the model.
        strict : bool, optional
            If True, enforces model scope to be found in the tag.
        recursive : bool, optional
            Whether the search for the model scope element should be recursive.
            Default is True.

        Returns
        -------
        Self | None
            An instance of the model if found, otherwise None.

        Raises
        ------
        ModelNotFoundException
            If the model's scope is not found and strict is True.
        FieldExtractionException
            If any model field failed to be extracted.
        """
        bound = cls.scope.find(tag=tag, strict=False, recursive=recursive)

        if bound is not None:
            return cls._find(bound)

        if strict:
            raise exc.ModelNotFoundException(
                f"Scope for the model '{cls.__name__}' was not found "
                f"by '{cls.scope}' in tag."
            )

        return None

    @classmethod
    def _find(cls, tag: Tag) -> Self:
        """
        Internal method to find and initialize a model instance from a given tag.
        By default, perform recursive, non-strict search for model fields.

        Parameters
        ----------
        tag : Tag
            The tag within which to search for model fields (scope element).

        Returns
        -------
        Self
            An instance of the model.

        Raises
        ------
        FieldExtractionException
            If any model field failed to be extracted.
        """
        params = {}

        for key, selector in cls.fields.items():
            try:
                result = selector.find(
                    tag=tag,
                    strict=c.DEFAULT_STRICT,
                    recursive=c.DEFAULT_RECURSIVE,
                )
            except exc.RequiredConstraintException as e:
                raise exc.FieldExtractionException(
                    f"Field '{key}' is required and was not found in model '{cls.__name__}' "
                    f"for element:\n{tag.prettify()}"
                ) from e
            except TagSearcherExceptions as e:
                raise exc.FieldExtractionException(
                    f"Extracting field '{key}' failed in model '{cls.__name__}' "
                    f"for element:\n{tag.prettify()}"
                ) from e

            params[key] = result

        return cls(**params)

    @classmethod
    def find_all(
        cls,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Self]:
        """
        Searches for and returns all instances of the model within the provided tag.
        By default, perform recursive, non-strict search for model fields
        just like in `find` method.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to search within for the model.
        recursive : bool, optional
            Whether the search for the model scope element should be recursive.
            Default is True.
        limit : int, optional
            Maximum number of model instances to return. Default is None, which
            returns all instances found.

        Returns
        -------
        list[Self]
            A list of model instances found within the tag.
        """
        elements = cls.scope.find_all(tag=tag, recursive=recursive, limit=limit)
        return [cls._find(element) for element in elements]

    def migrate(
        self,
        model: Type[T],
        mapping: Optional[dict[Type[BaseModel], Union[Type, MigrationSchema]]] = None,
        **kwargs,
    ) -> T:
        """
        Migrates the model instance to another model class using its fields
        in target class initialization. Recursively migrates nested models, creating
        new instances, even when target model is not defined in the mapping.

        Parameters
        ----------
        model : Type[Model]
            The target model class to migrate the instance to.
        mapping : dict[Type[BaseModel], Union[Type, MigrationSchema]], optional
            Mapping of base model fields to target models. By default, if field
            is instance of `BaseModel`, it will be passed directly to the target model.
        kwargs : Any
            Additional keyword arguments to pass to model initialization.

        Migrating to the same model is equivalent to creating deep copy of the model,
        which can be achieved by calling `copy` method.

        Returns
        -------
        Model
            An instance of the target model class.
        """
        mapping = mapping or {}

        include = {key for key, value in self.__class__.fields.items() if value.migrate}
        params = self.attributes.copy()

        for name, value in self.attributes.items():
            if name not in include:
                params.pop(name)

            if not isinstance(value, BaseModel):
                continue

            schema = mapping.get(value.__class__, value.__class__)

            if isinstance(schema, MigrationSchema):
                params[name] = value.migrate(
                    schema.target,
                    mapping=mapping,
                    **schema.params,
                )
                continue

            params[name] = value.migrate(schema, mapping=mapping)

        return model(**params, **kwargs)

    def copy(self) -> Self:
        """
        Creates a deep copy of the model instance by migrating
        it to the same model class.
        Only model fields defined in `attributes` are used to create new instance.

        Returns
        -------
        Self
            A deep copy of the model instance.
        """
        return self.migrate(self.__class__)

    def __setattr__(self, key, value) -> None:
        initialized = getattr(self, c.INITIALIZED, False)

        if not initialized:
            return super().__setattr__(key, value)

        fields = set(self.attributes.keys())

        if key not in fields:
            raise AttributeError(
                f"Cannot set attribute '{key}'. Only fields - {fields} - are allowed."
            )

        if self.__class__.__frozen__:
            raise exc.FrozenModelException(
                f"Model '{self.__class__}' is frozen and attributes of its instance "
                "cannot be modified."
            )

        super().__setattr__(key, value)

    def __hash__(self) -> int:
        if not self.__class__.__frozen__:
            raise TypeError(
                f"Cannot hash instance of model {self.__class__}, which is not frozen."
            )

        include = {key for key, value in self.__class__.fields.items() if value.compare}
        # Compute hash based on the model's attributes hashes and the class itself
        field_hashes = (
            hash((key, value))
            for key, value in self.attributes.items()
            if key in include
        )
        return hash((self.__class__, tuple(field_hashes)))

    def __repr__(self) -> str:
        include = {key for key, value in self.__class__.fields.items() if value.repr}
        params = [
            f"{name}={value!r}"
            for name, value in self.attributes.items()
            if name in include
        ]
        return f"{self.__class__.__name__}({', '.join(params)})"

    def __eq__(self, x: Any) -> bool:
        """
        Checks if provided object is equal to the model instance.
        They need to be of the same class and have the same field values.
        """
        if not isinstance(x, self.__class__):
            return False

        include = {key for key, value in self.__class__.fields.items() if value.compare}
        fields = self.attributes.keys()
        return all(
            getattr(self, key) == getattr(x, key) for key in fields if key in include
        )



"""
Module provides `nth-of-type` selector implementations for `SoupSelector`.
It allows you to search for the nth occurrence of an element,
similar to how the CSS `nth-of-type` pseudo-class works.
However, instead of being limited to css selectors, it works with any `SoupSelector` instance.

Classes
-------
`NthOfSelector` - Selects nth element matching given selector
`NthLastOfSelector` - Selects nth last element matching given selector
`OnlyOfSelector` - Selects only element matching given selector
"""

from .nth_soup_selector import NthLastOfSelector, NthOfSelector, OnlyOfSelector

__all__ = ["NthLastOfSelector", "NthOfSelector", "OnlyOfSelector"]



"""
Module for operations specific to BeautifulSoup tags.

Classes
-------
- `Text` - Extracts text from `bs4.Tag` - most common operation.
- `Href` - Extracts href attribute from `bs4.Tag`
- `Parent` - Extracts parent of `bs4.Tag`

These components are design to be used for processing html tags and extracting
desired information. They can be used in combination with selectors.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union, overload

from bs4 import Tag

from soupsavvy.base import BaseOperation, OperationSearcherMixin, SoupSelector

if TYPE_CHECKING:
    from soupsavvy.operations.general import OperationPipeline
    from soupsavvy.selectors.logical import SelectorList


class Text(OperationSearcherMixin):
    """
    Operation to extract text from `bs4.Tag`.
    Wrapper of most common operation used in web scraping.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... operation = Text()
    ... operation.execute(tag)
    "Extracted text from the tag"

    Wrapper for `bs4.Tag.get_text` method, that exposes all its options,
    and imitates its default behavior.

    Implements `TagSearcher` interface for convenience. It has find methods
    that can be used to extract text from provided tag.

    Example
    -------
    >>> from soupsavvy.operations import Text
    ... operation = Text()
    ... operation.find(tag)
    "Text"
    """

    def __init__(self, separator: str = "", strip: bool = False) -> None:
        """
        Initializes `Text` operation with optional separator and strip options.

        Parameters
        ----------
        separator : str, optional
            Separator used to join strings from multiple text nodes, by default "".
        strip : bool, optional
            Flag to strip the text nodes from whitespaces and newline characters,
            by default False.
        """
        self.separator = separator
        self.strip = strip

    def _execute(self, arg: Tag) -> str:
        """Extracts text from `bs4.Tag`."""
        return arg.get_text(separator=self.separator, strip=self.strip)

    def __eq__(self, x: Any) -> bool:
        # equal only if separator and strip are the same
        if not isinstance(x, Text):
            return False

        return self.separator == x.separator and self.strip == x.strip

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(separator={self.separator}, strip={self.strip})"
        )


class Href(OperationSearcherMixin):
    """
    Operation to extract href attribute from `bs4.Tag`.
    Wrapper of one of the common operation used in web scraping.
    If href attribute is not present, returns None.

    Example
    -------
    >>> from soupsavvy.operations import Href
    ... operation = Href()
    ... operation.execute(tag)
    "www.example.com"

    Implements `TagSearcher` interface for convenience. It has find methods
    that can be used to extract href from provided tag.

    Example
    -------
    >>> from soupsavvy.operations import Href
    ... operation = Href()
    ... operation.find(tag)
    "www.example.com"
    """

    def _execute(self, arg: Tag) -> Optional[str]:
        """
        Extracts href attribute from a BeautifulSoup Tag.
        If attribute is not present, returns None.
        """
        return arg.get_attribute_list("href")[0]

    def __eq__(self, x: Any) -> bool:
        # equal only if both are instances of Href
        return isinstance(x, Href)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class Parent(BaseOperation, SoupSelector):
    """
    Operation to extract parent of `bs4.Tag`.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent()
    ... operation.execute(tag)
    "<div>...</div>"

    Implements `SoupSelector` interface for convenience and can be used
    to extract parent of a provided tag without any conditions.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent()
    ... operation.find(tag)
    "<div>--tag--</div>"

    `Parent` has `BaseOperation` higher in MRO than `SoupSelector`, so, using pipe
    operator `|` on `Parent` object will result in `OperationPipeline` instance.

    Example
    -------
    >>> from soupsavvy.operations import Parent
    ... operation = Parent() | Parent()
    ... operation.execute(tag)
    "<div><div>--tag--</div></div>"

    Notes
    -----
    If element does not have parent, returns None.
    """

    def _execute(self, arg: Tag) -> Optional[Tag]:
        """Extracts parent of `bs4.Tag`."""
        return arg.parent

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        return [self.execute(tag)]

    def __eq__(self, x: Any) -> bool:
        # equal only if both are instances of Parent
        return isinstance(x, Parent)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    @overload  # type: ignore
    def __or__(self, other: BaseOperation) -> OperationPipeline: ...

    @overload
    def __or__(self, other: SoupSelector) -> SelectorList: ...

    def __or__(
        self, other: Union[BaseOperation, SoupSelector]
    ) -> Union[OperationPipeline, SelectorList]:
        if isinstance(other, SoupSelector):
            return SoupSelector.__or__(self, other)

        return BaseOperation.__or__(self, other)



"""
Module with selection pipeline class.
Pipeline for chaining selector and operation together, used as a bridge between
selecting html elements and processing the data.
"""

from __future__ import annotations

from typing import Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.base import BaseOperation, check_operation
from soupsavvy.interfaces import Comparable, TagSearcher


class SelectionPipeline(TagSearcher, Comparable):
    """
    Class for chaining searcher and operation together.
    Uses searcher to find information in tag and operation to process the data.

    Example
    -------
    >>> from soupsavvy import TypeSelector
    ... from soupsavvy.operations import Operation, Text
    ... pipeline = TypeSelector("span") | Text()
    ... pipeline.find(soup)
    'information'

    Most common way of creating a pipeline is using the `|` operator
    on selector and operation.
    """

    def __init__(self, selector: TagSearcher, operation: BaseOperation) -> None:
        """
        Initializes `SelectionPipeline` with selector and operation.

        Parameters
        ----------
        selector : TagSearcher
            Selector used for finding target information in the tag.
        operation : BaseOperation
            Operation used for processing the data.
        """
        if not isinstance(selector, TagSearcher):
            raise exc.NotTagSearcherException(
                f"Expected {TagSearcher.__name__}, got {type(selector)}"
            )

        self._selector = selector
        self._operation = operation

    @property
    def selector(self) -> TagSearcher:
        """
        Returns `TagSearcher` object of this pipeline used for finding target
        information in the tag.

        Returns
        -------
        TagSearcher
            TagSearcher object used in this pipeline.
        """
        return self._selector

    @property
    def operation(self) -> BaseOperation:
        """
        Returns `BaseOperation` object of this pipeline used for processing the data.

        Returns
        -------
        BaseOperation
            BaseOperation object used in this pipeline.
        """
        return self._operation

    def find(
        self,
        tag: Tag,
        strict: bool = False,
        recursive: bool = True,
    ) -> Any:
        """
        Finds a first element matching selector and processes it with operation.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        strict : bool, optional
            If True, enforces results to be found in the tag, by default False.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.

        Returns
        -------
        Any
            Result of the operation applied to the found tag.

        Raises
        ------
        TagNotFoundException
            If strict parameter is set to `True` and none matching tag was found.
        """
        return self.operation.execute(
            self.selector.find(tag, strict=strict, recursive=recursive)
        )

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Finds all elements matching selector and processes them with operation.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` object to process.
        recursive : bool, optional
            Specifies if search should be recursive.
            If set to `False`, only direct children of the tag will be searched.
            By default `True`.
        limit : int, optional
            Specifies maximum number of results to return in a list.
            By default `None`, everything is returned.

        Returns
        -------
        list[Any]
            A list of results, if none found, the list is empty.
        """
        return [
            self.operation.execute(element)
            for element in self.selector.find_all(tag, recursive=recursive, limit=limit)
        ]

    def __or__(self, x: Any) -> SelectionPipeline:
        """
        Overrides `__or__` method called also by pipe operator '|'.
        Creates new `SelectionPipeline` by extending operations with provided one.

        Parameters
        ----------
        x : BaseOperation
            `BaseOperation` object used to extend the pipeline.

        Returns
        -------
        SelectionPipeline
            New `SelectionPipeline` with extended operations.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        x = check_operation(x)
        operation = self.operation | x
        return SelectionPipeline(selector=self.selector, operation=operation)

    def __eq__(self, x) -> bool:
        # equal only if both selector and operation are the same
        if not isinstance(x, SelectionPipeline):
            return False

        return self.selector == x.selector and self.operation == x.operation

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.selector}, {self.operation})"



"""
Module with wrappers for selectors used mostly in model schemas.
Wrappers are used to control behavior and define how fields of the models are defined.

Classes
-------
- `All` - Wrapper to find all information matching criteria.
- `Default` - Wrapper to set default value if information is not found.
- `Required` - Wrapper to enforce that information must be found.
"""

from typing import Any, Optional

from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.base import check_operation
from soupsavvy.interfaces import Comparable, TagSearcher
from soupsavvy.operations.selection_pipeline import SelectionPipeline


class FieldWrapper(TagSearcher, Comparable):
    """
    A wrapper for `TagSearcher` objects, that acts as a higher order searcher,
    which controls behavior of the wrapped searcher.
    Used as field to defined model.
    """

    def __init__(self, selector: TagSearcher) -> None:
        """
        Initializes wrapper with a `TagSearcher` instance.

        Parameters
        ----------
        selector : TagSearcher
            The `TagSearcher` instance to be wrapped.

        Raises
        ------
        NotTagSearcherException
            If provided object is not an instance of `TagSearcher`.
        """

        if not isinstance(selector, TagSearcher):
            raise exc.NotTagSearcherException(
                f"Expected {TagSearcher.__name__}, got {type(selector)}."
            )

        self._selector = selector

    @property
    def selector(self) -> TagSearcher:
        """
        Returns searcher, that is wrapped by this wrapper.

        Returns
        -------
        TagSearcher
            `TagSearcher` instance wrapped by this wrapper.
        """
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Any]:
        """
        Find all matching tags using the wrapped selector.
        Used for compatibility with `TagSearcher` interface,
        delegates to wrapped selector.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` to search within.
        recursive : bool, optional
            Whether to search recursively, by default True.
        limit : int, optional
            Limit the number of results, by default None.

        Returns
        -------
        list[Any]
            A list of found results.
        """
        return self.selector.find_all(tag, recursive=recursive, limit=limit)

    def __eq__(self, x: Any) -> bool:
        """
        Check if two `FieldWrapper` instances are equal.
        They need to be of the same class and wrap the same selector.

        Parameters
        ----------
        x : Any
            The object to compare with.

        Returns
        -------
        bool
            True if the instances are equal, False otherwise.
        """
        return isinstance(x, self.__class__) and self.selector == x.selector

    def __or__(self, x: Any) -> SelectionPipeline:
        """
        Overrides `__or__` method called also by pipe operator '|'.
        Combines this `FieldWrapper` with an operation.

        Parameters
        ----------
        x : BaseOperation
            The operation to be combined with this `FieldWrapper`.

        Returns
        -------
        SelectionPipeline
            New `SelectionPipeline` object created from
            combining selector and operation.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        x = check_operation(x)
        return SelectionPipeline(selector=self, operation=x)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector})"

    def __repr__(self) -> str:
        return str(self)


class All(FieldWrapper):
    """
    Field wrapper for selecting multiple elements matching the selector.
    Forces find method to fall back to find_all method and return all matches.

    Example
    -------
    >>> from soupsavvy.models import All
    ... from soupsavvy import TypeSelector
    ... selector = All(TypeSelector("div"))
    ... selector.find(tag)
    [element1, element2, element3]
    """

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True) -> list[Any]:
        """
        Find all matching tags using the wrapped selector,
        enforcing the use of find_all method.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` to search within.
        strict : bool, optional
            Ignored, as this method always falls back to `find_all`.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        list[Any]
            A list of matching results.
        """
        return self.selector.find_all(tag, recursive=recursive)


class Required(FieldWrapper):
    """
    Field wrapper for enforcing matched element not to be None.
    Raises an exception if searcher does not find any matches.

    Example
    -------
    >>> from soupsavvy.models import Required
    ... from soupsavvy import TypeSelector
    ... selector = Required(TypeSelector("div"))
    ... selector.find(tag)
    RequiredConstraintException
    """

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True) -> Any:
        """
        Finds a required element using the wrapped selector,
        enforcing matched element not to be None.
        If any exception is raised during the search, it's propagated to the caller.

        Parameters
        ----------
        tag : Tag
            Any `bs4.Tag` to search within.
        strict : bool, optional
            If True, raises an exception if no matches are found, by default False.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        Any
            The found element.

        Raises
        ------
        RequiredConstraintException
            If selector returns None, indicating that required element was not found.
        """
        result = self.selector.find(tag=tag, strict=strict, recursive=recursive)

        if result is None:
            raise exc.RequiredConstraintException("Required element was not found")

        return result


class Default(FieldWrapper):
    """
    Field wrapper for returning a default value if no match is found.

    Example
    -------
    >>> from soupsavvy.models import Default
    ... from soupsavvy import TypeSelector
    ... selector = Default(TypeSelector("div"), default="1234")
    ... selector.find(tag)
    "1234"
    """

    def __init__(self, selector: TagSearcher, default: Any) -> None:
        """
        Initializes `Default` field wrapper.

        Parameters
        ----------
        selector : TagSearcher
            The `TagSearcher` instance to be wrapped.
        default : Any
            The default value to return if no match is found.
        """
        super().__init__(selector)
        self.default = default

    def find(self, tag: Tag, strict: bool = False, recursive: bool = True):
        """
        Finds an element, returning a default value if None was returned
        by wrapped selector. Any exception raised during the search is propagated.

        Parameters
        ----------
        tag : Tag
            Any bs4.Tag` to search within.
        strict : bool, optional
            If True, raises an exception if no matches are found, by default False.
        recursive : bool, optional
            Whether to search recursively, by default True.

        Returns
        -------
        Any
            The found element or the default value if not found.
        """
        result = self.selector.find(tag=tag, strict=strict, recursive=recursive)
        return self.default if result is None else result

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.selector}, default={self.default})"



"""
Module for general purpose operations.

Classes
-------
- `Operation` - Executes custom operation with any function.
- `OperationPipeline` - Chains multiple operations together.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import soupsavvy.exceptions as exc
from soupsavvy.base import BaseOperation, OperationSearcherMixin, check_operation


class OperationPipeline(OperationSearcherMixin):
    """
    Pipeline for chaining multiple operations together.
    Applies each operation in sequence, passing the result to the next.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... pipeline = Operation(int) | Operation(lambda x: x + 1)
    ... pipeline.execute("1")
    2

    Most common way of creating a pipeline is using the `|` operator
    on two operations.

    `OperationPipeline` is operation-searcher mixin, which means it can be used
    to find information in `bs4.Tag` object directly with find methods.
    This way, it can be used as field in model or `execute` method can be replaced
    with `find` method, which would produce the same result.
    """

    def __init__(
        self,
        operation1: BaseOperation,
        operation2: BaseOperation,
        /,
        *operations: BaseOperation,
    ) -> None:
        """
        Initializes `OperationPipeline` with multiple operations.

        Parameters
        ----------
        operations : BaseOperation
            `BaseOperation` instances to be chained together.

        Raises
        ------
        NotOperationException
            If any of the input operations is not an instance of `BaseOperation`.
        """
        self.operations = [
            check_operation(operation)
            for operation in (operation1, operation2, *operations)
        ]

    def _execute(self, arg: Any) -> Any:
        """
        Executes each operation in sequence, passing the result to the next one.
        If operation is breaking, returns the result immediately.
        """
        for operation in self.operations:

            try:
                arg = operation.execute(arg)
            except exc.BreakOperationException as e:
                return e.result

        return arg

    def __or__(self, x: Any) -> OperationPipeline:
        """
        Overrides `__or__` method called also by pipe operator '|'.
        Creates new `OperationPipeline` object extended by provided operation.

        Parameters
        ----------
        x : BaseOperation
            `BaseOperation` object used to extend the pipeline.

        Returns
        -------
        OperationPipeline
            New `OperationPipeline` with extended operations.

        Raises
        ------
        NotOperationException
            If provided object is not an instance of `BaseOperation`.
        """
        x = check_operation(x)
        return OperationPipeline(*self.operations, x)

    def __eq__(self, x: Any) -> bool:
        # equal only if operations are the same
        if not isinstance(x, OperationPipeline):
            return False

        return self.operations == x.operations

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.operations})"


class Operation(OperationSearcherMixin):
    """
    Custom operation that wraps any function
    to be used with other `soupsavvy` components.

    Example
    -------
    >>> from soupsavvy.operations import Operation
    ... operation = Operation(str.lower)
    ... operation.execute("TEXT")
    "text"

    `Operation` is operation-searcher mixin, which means it can be used
    to find information in `bs4.Tag` directly with find methods.
    This way, it can be used as field in model or `execute` method can be replaced
    with `find` method, which would produce the same result.
    """

    def __init__(self, func: Callable, *args, **kwargs) -> None:
        """
        Initializes `Operation` with provided function and optional arguments.

        Parameters
        ----------
        func : Callable
            Any callable object that can be called with one positional argument.
        *args : Any
            Additional positional arguments passed to the operation function.
        **kwargs : Any
            Additional keyword arguments passed to the operation function.
        """
        self.operation = func
        self.args = args
        self.kwargs = kwargs

    def _execute(self, arg: Any) -> Any:
        """Executes the custom operation."""
        return self.operation(arg, *self.args, **self.kwargs)

    def __eq__(self, x: Any) -> bool:
        # functions used as operations need to be the same object
        # and have the same function arguments if provided
        if not isinstance(x, Operation):
            return False

        return (
            self.operation is x.operation
            and self.args == x.args
            and self.kwargs == x.kwargs
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.operation})"



"""
Package with models for defining search schemas with operations and selectors.

Classes
-------
- `BaseModel` - Base class for defining search schemas.
- `Field` - Field configuration for search schema.
- `MigrationSchema` - Configuration of model migration.
- `All` - Wrapper to find all information matching criteria.
- `Default` - Wrapper to set default value if information is not found.
- `Required` - Wrapper to enforce that information must be found.
"""

from .base import BaseModel, Field, MigrationSchema, post
from .wrappers import All, Default, Required

__all__ = [
    "BaseModel",
    "All",
    "Default",
    "Required",
    "post",
    "Field",
    "MigrationSchema",
]



"""
Package with classes for searching based on CSS selectors.

Contains implementation of basic CSS pseudo-classes like
:only-child, :empty, :nth-child().

They can be used in combination with other `SoupSelector` objects
to create more complex search procedures.

Classes
-------
- OnlyChild
- Empty
- FirstChild
- LastChild
- NthChild
- NthLastChild
- FirstOfType
- LastOfType
- NthOfType
- NthLastOfType
- OnlyOfType
- CSS - wrapper for simple search with CSS selectors
"""

from .selectors import (
    CSS,
    Empty,
    FirstChild,
    FirstOfType,
    LastChild,
    LastOfType,
    NthChild,
    NthLastChild,
    NthLastOfType,
    NthOfType,
    OnlyChild,
    OnlyOfType,
)

__all__ = [
    "CSS",
    "Empty",
    "FirstChild",
    "FirstOfType",
    "LastChild",
    "LastOfType",
    "NthChild",
    "NthLastChild",
    "NthLastOfType",
    "NthOfType",
    "OnlyChild",
    "OnlyOfType",
]
